# coding: utf-8
"""
自動インサイト生成のテスト

InsightGenerator の全機能を検証：
- インサイト自動生成
- パターン検出（トピック、トレンド、異常値）
- フレーズクラスタリング
- レポート生成
"""

import pytest
import pandas as pd
import tempfile
import os
from japhrase import InsightGenerator, Insight


class TestInsightDataclass:
    """Insight データクラスのテスト"""

    def test_insight_creation(self):
        """Insight オブジェクトの作成"""
        insight = Insight(
            category='dominant_topic',
            title='ドミナントトピック',
            description='重要なトピックです',
            phrases=[('テスト', 0.5), ('実装', 0.3)],
            confidence=0.85
        )

        assert insight.category == 'dominant_topic', "カテゴリ"
        assert insight.title == 'ドミナントトピック', "タイトル"
        assert len(insight.phrases) == 2, "フレーズリスト"
        assert insight.confidence == 0.85, "確信度"


class TestDetectDominantTopics:
    """ドミナントトピック検出のテスト"""

    def test_detect_dominant_topics_basic(self):
        """基本的なドミナントトピック検出"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', 'テスト', 'テスト', '実装', '実装', 'バグ'],
            'freq': [100, 50, 30, 80, 40, 20],
        })

        insights = generator.generate_insights(df_phrases)

        # ドミナントトピックインサイトを探す
        dominant = next((i for i in insights if i.category == 'dominant_topic'), None)

        assert dominant is not None, "ドミナントトピックが検出"
        assert len(dominant.phrases) == 5, "TOP 5 フレーズ"
        assert dominant.phrases[0][0] == 'テスト', "最高頻度フレーズ"

    def test_dominant_topic_confidence(self):
        """ドミナントトピックの確信度"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト'] * 100,
            'freq': [1] * 100,
        })

        insights = generator.generate_insights(df_phrases)

        dominant = next((i for i in insights if i.category == 'dominant_topic'), None)

        # 1フレーズが100%占める
        assert dominant.confidence == 0.95, "高い確信度"


class TestDetectEmergingThemes:
    """浮上テーマ検出のテスト"""

    def test_detect_emerging_themes_basic(self):
        """基本的な浮上テーマ検出"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['古い', '古い', '古い', '新しい', '新しい'],
            'freq': [3, 2, 1, 2, 1],
        })

        texts = [
            '古いテキスト1',
            '古いテキスト2',
            '古いテキスト3',
            '新しいテキスト1',
            '新しいテキスト2',
            '新しいテキスト3',
        ]

        insights = generator.generate_insights(df_phrases, texts)

        # テキストが少なくても処理可能
        assert len(insights) >= 1, "インサイトが生成"

    def test_no_emerging_themes_small_dataset(self):
        """テキストが少ない場合"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', '実装'],
            'freq': [5, 3],
        })

        texts = ['テキスト1', 'テキスト2']

        insights = generator.generate_insights(df_phrases, texts)

        # 浮上テーマは検出されない（テキスト < 10）
        emerging = next((i for i in insights if i.category == 'emerging_theme'), None)

        assert emerging is None, "テキスト少数では浮上テーマ検出されない"

    def test_emerging_themes_growth_detection(self):
        """成長率の検出"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['新規', '成熟'],
            'freq': [2, 1],
        })

        texts = [
            '成熟テキスト1', '成熟テキスト2', '成熟テキスト3',
            '成熟テキスト4', '成熟テキスト5',
            '新規テキスト1', '新規テキスト2', '新規テキスト3',
            '新規テキスト4', '新規テキスト5',
            '新規テキスト6', '新規テキスト7',
        ]

        insights = generator.generate_insights(df_phrases, texts)

        # インサイトが生成される
        assert len(insights) > 0, "インサイト生成"


class TestDetectAnomalies:
    """異常値検出のテスト"""

    def test_detect_anomalies_without_stats(self):
        """統計スコアなしでの異常検出"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', 'テスト', '実装'],
            'freq': [100, 80, 50],
        })

        insights = generator.generate_insights(df_phrases, statistical_scores=None)

        # 統計スコアなしでは異常は検出されない
        anomalies = [i for i in insights if i.category == 'anomaly']

        assert len(anomalies) == 0, "統計スコアなしでは異常検出なし"

    def test_detect_anomalies_with_stats(self):
        """統計スコア付きでの異常検出"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', '実装', 'バグ'],
            'freq': [100, 80, 50],
        })

        statistical_scores = pd.DataFrame({
            'phrase': ['テスト', '実装', 'バグ'],
            'frequency': [100, 80, 50],
            'is_significant': [False, True, True],
        })

        insights = generator.generate_insights(
            df_phrases,
            statistical_scores=statistical_scores
        )

        # 'テスト' が有意でないが頻出なので異常と判定される可能性
        assert len(insights) > 0, "インサイト生成"


class TestAnalyzeNoiseRatio:
    """ノイズ比率分析のテスト"""

    def test_analyze_noise_ratio_high_noise(self):
        """ノイズが多い場合"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['a', 'b', '、', '。', 'テスト', '実装'],
            'freq': [100, 80, 60, 50, 30, 20],
        })

        insights = generator.generate_insights(df_phrases)

        noise = next((i for i in insights if i.category == 'noise_analysis'), None)

        # ノイズが5%以上 → インサイト生成
        if noise is not None:
            assert noise.confidence > 0.5, "ノイズ比率が高い"

    def test_analyze_noise_ratio_low_noise(self):
        """ノイズが少ない場合"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', '実装', 'バグ', '修正', '改善'],
            'freq': [100, 80, 60, 50, 40],
        })

        insights = generator.generate_insights(df_phrases)

        noise = next((i for i in insights if i.category == 'noise_analysis'), None)

        assert noise is None, "ノイズ少ない場合は表示されない"


class TestEvaluateTextQuality:
    """テキスト品質評価のテスト"""

    def test_evaluate_text_quality_low(self):
        """品質が低い（語彙多様性低い）"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト'] * 100,
            'freq': [1] * 100,
        })

        insights = generator.generate_insights(df_phrases)

        quality = next((i for i in insights if i.category == 'text_quality'), None)

        assert quality is not None, "品質評価が生成"
        # Gini 係数が低い（語彙少ない）
        if 'Gini' in str(quality.description) or '低' in quality.description:
            assert '低' in quality.description, "低品質と判定"

    def test_evaluate_text_quality_high(self):
        """品質が高い（語彙多様性高い）"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': [f'フレーズ{i}' for i in range(100)],
            'freq': [1] * 100,
        })

        insights = generator.generate_insights(df_phrases)

        quality = next((i for i in insights if i.category == 'text_quality'), None)

        assert quality is not None, "品質評価が生成"


class TestGenerateRecommendations:
    """推奨アクション生成のテスト"""

    def test_generate_recommendations_with_noise(self):
        """ノイズ推奨"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['a', 'b', '、', 'テスト', '実装'],
            'freq': [100, 80, 60, 50, 40],
        })

        insights = generator.generate_insights(df_phrases)

        recommendations = [i for i in insights if i.category == 'recommendation']

        # ノイズが多い場合は推奨が生成される
        if len([i for i in insights if i.category == 'noise_analysis']) > 0:
            assert len(recommendations) > 0, "推奨が生成"


class TestClusterRelatedPhrases:
    """関連フレーズクラスタリングのテスト"""

    def test_cluster_similar_phrases(self):
        """類似フレーズのクラスタリング"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', 'テスツ', 'てすと', '実装', '実装'],
            'freq': [100, 50, 30, 80, 20],
        })

        clusters = generator.cluster_related_phrases(df_phrases, similarity_threshold=0.7)

        assert isinstance(clusters, dict), "辞書を返す"
        assert len(clusters) >= 2, "複数クラスタ"

    def test_cluster_exact_match(self):
        """同一フレーズのクラスタリング"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', 'テスト', '実装'],
            'freq': [100, 50, 80],
        })

        clusters = generator.cluster_related_phrases(df_phrases, similarity_threshold=0.5)

        # 'テスト' は1つのクラスタになる
        test_clusters = [v for v in clusters.values() if 'テスト' in v]

        assert len(test_clusters) > 0, "テストをクラスタ"

    def test_cluster_threshold(self):
        """閾値による影響"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['A', 'AB', 'ABC', 'ABCD'],
            'freq': [1, 1, 1, 1],
        })

        clusters_strict = generator.cluster_related_phrases(df_phrases, similarity_threshold=0.9)
        clusters_loose = generator.cluster_related_phrases(df_phrases, similarity_threshold=0.5)

        # 厳しい閾値 → より多くのクラスタ
        assert len(clusters_strict) >= len(clusters_loose), "厳しい閾値でクラスタ増加"


class TestExportInsights:
    """インサイトエクスポートのテスト"""

    def test_export_insights_csv(self):
        """CSV 出力"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', 'テスト', '実装', '実装'],
            'freq': [100, 50, 80, 30],
        })

        insights = generator.generate_insights(df_phrases)

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.csv') as f:
            filepath = f.name

        try:
            generator.export_insights(filepath)

            assert os.path.exists(filepath), "ファイルが生成"
            assert os.path.getsize(filepath) > 0, "ファイルに内容"

            # CSV として読み込み可能か
            df = pd.read_csv(filepath)
            assert len(df) > 0, "CSV が読み込める"
        finally:
            if os.path.exists(filepath):
                os.unlink(filepath)

    def test_export_insights_no_insights(self):
        """インサイトなしでのエクスポート"""
        generator = InsightGenerator()

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.csv') as f:
            filepath = f.name

        try:
            # インサイトなし
            generator.insights = []
            generator.export_insights(filepath)

            # ファイルは生成されない（または空で警告が出る）
            if os.path.exists(filepath):
                # ファイルが空の場合はパスする
                try:
                    df = pd.read_csv(filepath)
                except pd.errors.EmptyDataError:
                    # 空ファイルは正常なケース
                    pass
        finally:
            if os.path.exists(filepath):
                os.unlink(filepath)


class TestGenerateReport:
    """レポート生成のテスト"""

    def test_generate_report_basic(self):
        """基本的なレポート生成"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', 'テスト', '実装', '実装', '実装'],
            'freq': [100, 50, 80, 30, 20],
        })

        insights = generator.generate_insights(df_phrases)

        report = generator.generate_report()

        assert isinstance(report, str), "文字列を返す"
        assert '自動インサイトレポート' in report, "タイトル"
        assert '生成日時' in report, "日時"
        assert 'インサイト数' in report, "インサイト数"

    def test_generate_report_content(self):
        """レポート内容の検証"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト', 'テスト', '実装'],
            'freq': [100, 50, 80],
        })

        insights = generator.generate_insights(df_phrases)

        report = generator.generate_report()

        # 各インサイトの情報が含まれるか
        for insight in insights:
            if len(report) > 500:  # レポート内容が充実している場合
                # タイトルやカテゴリが含まれる可能性
                assert 'カテゴリ' in report or insight.title in report, "インサイト情報"

    def test_report_formatting(self):
        """レポートのフォーマット"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['A', 'A', 'B'],
            'freq': [100, 50, 80],
        })

        insights = generator.generate_insights(df_phrases)

        report = generator.generate_report()

        # 基本的なレポート構造
        lines = report.split('\n')
        assert len(lines) > 1, "複数行のレポート"


class TestIntegration:
    """統合テスト"""

    def test_full_insight_pipeline(self):
        """インサイト生成の全フロー"""
        generator = InsightGenerator()

        # テストデータ
        df_phrases = pd.DataFrame({
            'seqchar': [
                'AI', 'AI', 'AI', 'AI', 'AI',
                '機械学習', '機械学習', '機械学習',
                'テスト', 'テスト',
                'a', 'b', '、',  # ノイズ
            ],
            'freq': [100, 80, 60, 40, 20, 80, 50, 30, 50, 30, 100, 80, 60],
        })

        texts = [
            'AI技術の発展',
            'AI と機械学習',
            '機械学習の応用',
            'テストが重要',
            'テスト実施',
        ] * 2

        # インサイト生成
        insights = generator.generate_insights(df_phrases, texts)

        # 検証
        assert len(insights) > 0, "インサイトが生成"

        # クラスタリング
        clusters = generator.cluster_related_phrases(df_phrases)
        assert len(clusters) > 0, "クラスタ生成"

        # レポート生成
        report = generator.generate_report()
        assert len(report) > 100, "詳細レポート"

    def test_insight_with_statistical_scores(self):
        """統計スコア付きインサイト"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['重要', '重要', 'ノイズ', 'ノイズ'],
            'freq': [100, 50, 40, 30],
        })

        statistical_scores = pd.DataFrame({
            'phrase': ['重要', '重要', 'ノイズ', 'ノイズ'],
            'frequency': [100, 50, 40, 30],
            'is_significant': [True, True, False, False],
        })

        insights = generator.generate_insights(
            df_phrases,
            texts=None,
            statistical_scores=statistical_scores
        )

        # インサイトが生成される
        assert len(insights) > 0, "インサイト生成"

    def test_large_phrase_set(self):
        """大規模フレーズセット"""
        generator = InsightGenerator()

        # 1000個のフレーズ
        df_phrases = pd.DataFrame({
            'seqchar': [f'フレーズ{i}' for i in range(1000)],
            'freq': [100 - i for i in range(1000)],
        })

        insights = generator.generate_insights(df_phrases)

        # 処理できるか
        assert len(insights) > 0, "大規模データ処理"

        # レポート生成
        report = generator.generate_report()
        assert len(report) > 50, "レポート生成成功"

    def test_edge_case_single_phrase(self):
        """単一フレーズ"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['テスト'],
            'freq': [100],
        })

        insights = generator.generate_insights(df_phrases)

        # エラーなく処理
        assert isinstance(insights, list), "インサイトリスト"

    def test_edge_case_all_same_frequency(self):
        """すべてのフレーズが同じ頻度"""
        generator = InsightGenerator()

        df_phrases = pd.DataFrame({
            'seqchar': ['A', 'B', 'C', 'D', 'E'],
            'freq': [50] * 5,
        })

        insights = generator.generate_insights(df_phrases)

        assert len(insights) > 0, "均等分布でもインサイト生成"



"""
設定ファイル管理モジュール

.japhrase.toml / .japhrase.yml から設定を読み込む
"""

import os
from pathlib import Path
from typing import Dict, Any, Optional
import logging

logger = logging.getLogger(__name__)


class JaphraseConfig:
    """japhrase 設定ファイル管理クラス"""

    # デフォルト設定ファイル名
    CONFIG_FILES = ['.japhrase.toml', '.japhrase.yml', 'japhrase.toml', 'japhrase.yml']

    def __init__(self, config_file: Optional[str] = None):
        """
        設定を初期化

        Args:
            config_file: 明示的に指定する設定ファイルパス
        """
        self.config_file = config_file or self._find_config_file()
        self.config: Dict[str, Any] = {}

        if self.config_file:
            self._load_config()
        else:
            logger.debug("設定ファイルが見つかりません")

    @staticmethod
    def _find_config_file() -> Optional[str]:
        """現在のディレクトリと親ディレクトリから設定ファイルを探索"""
        current_dir = Path.cwd()

        # 現在のディレクトリから5階層上まで探索
        for parent in [current_dir, *current_dir.parents[:5]]:
            for config_name in JaphraseConfig.CONFIG_FILES:
                config_path = parent / config_name
                if config_path.exists():
                    logger.debug(f"設定ファイルが見つかりました: {config_path}")
                    return str(config_path)

        return None

    def _load_config(self):
        """設定ファイルから設定を読み込む"""
        if not self.config_file:
            return

        try:
            if self.config_file.endswith('.toml'):
                self._load_toml()
            elif self.config_file.endswith('.yml') or self.config_file.endswith('.yaml'):
                self._load_yaml()
            else:
                logger.warning(f"不正な設定ファイル形式: {self.config_file}")
        except Exception as e:
            logger.error(f"設定ファイルの読み込みエラー: {e}")

    def _load_toml(self):
        """TOML形式の設定ファイルを読み込む"""
        try:
            import tomllib
        except ImportError:
            try:
                import tomli as tomllib
            except ImportError:
                logger.error("toml ライブラリがインストールされていません")
                return

        try:
            with open(self.config_file, 'rb') as f:
                self.config = tomllib.load(f)
                logger.info(f"設定ファイルを読み込みました: {self.config_file}")
        except Exception as e:
            logger.error(f"TOML ファイルの読み込みエラー: {e}")

    def _load_yaml(self):
        """YAML形式の設定ファイルを読み込む"""
        try:
            import yaml
        except ImportError:
            logger.error("PyYAML がインストールされていません")
            return

        try:
            with open(self.config_file, 'r', encoding='utf-8') as f:
                self.config = yaml.safe_load(f) or {}
                logger.info(f"設定ファイルを読み込みました: {self.config_file}")
        except Exception as e:
            logger.error(f"YAML ファイルの読み込みエラー: {e}")

    def get(self, key: str, default: Any = None) -> Any:
        """ネストされたキーから値を取得"""
        keys = key.split('.')
        value = self.config

        for k in keys:
            if isinstance(value, dict):
                value = value.get(k)
                if value is None:
                    return default
            else:
                return default

        return value if value is not None else default

    def get_extractor_params(self) -> Dict[str, Any]:
        """PhraseExtracter 用のパラメータを取得"""
        params = {}

        # グローバル設定
        if 'global' in self.config:
            if 'preset' in self.config['global']:
                params['preset'] = self.config['global']['preset']

        # 分析設定
        if 'analysis' in self.config:
            analysis = self.config['analysis']
            if 'min_count' in analysis:
                params['min_count'] = analysis['min_count']
            if 'max_length' in analysis:
                params['max_length'] = analysis['max_length']
            if 'min_length' in analysis:
                params['min_length'] = analysis['min_length']
            if 'threshold_originality' in analysis:
                params['threshold_originality'] = analysis['threshold_originality']
            if 'weight_freq' in analysis:
                params['weight_freq'] = analysis['weight_freq']
            if 'weight_len' in analysis:
                params['weight_len'] = analysis['weight_len']

        # フィルタ設定
        if 'filter' in self.config:
            filter_cfg = self.config['filter']
            if 'ignore' in filter_cfg:
                # ignore リストを文字列に変換
                ignore_list = filter_cfg['ignore']
                if isinstance(ignore_list, list):
                    params['removes'] = ''.join(ignore_list)
            if 'knowns' in filter_cfg:
                params['knowns'] = filter_cfg['knowns']
            if 'unnecessary' in filter_cfg:
                params['unnecesary'] = filter_cfg['unnecessary']

        return params

    def get_workflow_params(self) -> Dict[str, Any]:
        """ワークフロー用のパラメータを取得"""
        return self.config.get('workflow', {})

    def get_check_rules(self) -> Dict[str, Any]:
        """check コマンド用のルール設定を取得"""
        return self.config.get('check', {})

    def get_output_config(self) -> Dict[str, Any]:
        """出力設定を取得"""
        return self.config.get('output', {})

    def display_config(self) -> str:
        """現在の設定を表示用に整形"""
        if not self.config:
            return "設定ファイルが読み込まれていません"

        lines = []
        lines.append(f"設定ファイル: {self.config_file}")
        lines.append("=" * 70)

        for section, content in self.config.items():
            lines.append(f"\n[{section}]")
            if isinstance(content, dict):
                for key, value in content.items():
                    lines.append(f"  {key} = {value}")
            else:
                lines.append(f"  {content}")

        return "\n".join(lines)



"""
計量文学（Stylometry）分析モジュール

文体の定量的特徴（語彙の豊かさ、文字種の比率など）を分析します。
著者判定や文体分類に用いられる統計指標を計算します。

Phase 2 強化 (2026-03):
- Hapax Legomena Ratio / Dis Legomena Ratio
- Brunet's W / Honoré's R / Sichel's S / Simpson's D
- MATTR (Moving Average Type-Token Ratio)
- Vocabulary Growth Curve (Heaps' Law fitting)
"""

__author__ = "Takeshi SHIMIZU"
__copyright__ = "Copyright 2023-2026"

import re
import math
import numpy as np
from collections import Counter
from typing import Dict, List, Optional, Tuple
import logging

logger = logging.getLogger(__name__)


class StylometryAnalyzer:
    """
    文体の定量的特徴を分析

    以下の指標を計算します：
    - TTR (Type-Token Ratio): 基本的な語彙多様性
    - Yule's K: 長文に強い語彙多様性指標
    - 文字種比率: 漢字/ひらがな/カタカナ/その他の構成

    使用例:
        >>> analyzer = StylometryAnalyzer()
        >>> result = analyzer.analyze_vocabulary_richness(text)
        >>> print(f"TTR: {result['ttr']:.3f}")
    """

    def __init__(self):
        """初期化"""
        pass

    def analyze_vocabulary_richness(self, text: str) -> Dict:
        """
        語彙多様性指標（TTR, Yule's K）を計算

        Parameters:
            text (str): 入力テキスト

        Returns:
            Dict: 語彙多様性の指標
                {
                    'total_tokens_est': int,    # 総トークン数（推定）
                    'unique_types_est': int,    # ユニーク語彙数（推定）
                    'ttr': float,               # Type-Token Ratio (0-1)
                    'yules_k': float,           # Yule's K指標
                    'assessment': str           # 評価テキスト
                }
        """
        try:
            from .extracter import PhraseExtracter

            # 簡易トークナイズ（PhraseExtracterのロジック利用）
            extractor = PhraseExtracter(
                min_count=1, min_length=2, verbose=0
            )
            df = extractor.extract([text])

            if df.empty:
                return {
                    'total_tokens_est': 0,
                    'unique_types_est': 0,
                    'ttr': 0.0,
                    'yules_k': 0.0,
                    'assessment': '分析対象がありません',
                }

            # 列名の互換性処理
            freq_col = 'freq' if 'freq' in df.columns else 'frequency'

            # 総トークン数（推定）
            N = int(df[freq_col].sum())
            # ユニーク語彙数
            V = len(df)

            # TTR (Type-Token Ratio)
            ttr = V / N if N > 0 else 0.0

            # Yule's K
            # K = 10^4 * (Σ(freq²) - N) / N²
            # 値が大きいほど語彙が単調（繰り返しが多い）
            sum_freq_sq = (df[freq_col] ** 2).sum()
            yules_k = 10000 * (sum_freq_sq - N) / (N ** 2) if N > 0 else 0.0

            return {
                'total_tokens_est': N,
                'unique_types_est': V,
                'ttr': round(ttr, 4),
                'yules_k': round(yules_k, 2),
                'assessment': self._assess_richness(yules_k),
            }

        except ImportError:
            logger.warning("StylometryAnalyzer: PhraseExtracterが必要です")
            return {
                'total_tokens_est': 0,
                'unique_types_est': 0,
                'ttr': 0.0,
                'yules_k': 0.0,
                'assessment': 'PhraseExtracterが未導入',
            }

    def analyze_char_type_ratio(self, text: str) -> Dict:
        """
        文字種比率（文体指紋）を計算

        Parameters:
            text (str): 入力テキスト

        Returns:
            Dict: 文字種の比率と文体診断
                {
                    'kanji_ratio': float,      # 漢字の比率
                    'hiragana_ratio': float,   # ひらがなの比率
                    'katakana_ratio': float,   # カタカナの比率
                    'ascii_ratio': float,      # ASCII（英数字など）の比率
                    'other_ratio': float,      # その他の比率
                    'style_type': str          # 文体の分類
                }
        """
        counts = Counter()
        total = 0

        for char in text:
            if char.isspace():
                continue
            total += 1
            if re.match(r'[一-龠]', char):
                counts['kanji'] += 1
            elif re.match(r'[ぁ-ん]', char):
                counts['hiragana'] += 1
            elif re.match(r'[ァ-ヶ]', char):
                counts['katakana'] += 1
            elif re.match(r'[a-zA-Z0-9]', char):
                counts['ascii'] += 1
            else:
                counts['other'] += 1

        if total == 0:
            return {
                'kanji_ratio': 0.0,
                'hiragana_ratio': 0.0,
                'katakana_ratio': 0.0,
                'ascii_ratio': 0.0,
                'other_ratio': 0.0,
                'style_type': '分析対象がありません',
            }

        kanji_ratio = counts['kanji'] / total
        hiragana_ratio = counts['hiragana'] / total
        katakana_ratio = counts['katakana'] / total
        ascii_ratio = counts['ascii'] / total
        other_ratio = counts['other'] / total

        return {
            'kanji_ratio': round(kanji_ratio, 3),
            'hiragana_ratio': round(hiragana_ratio, 3),
            'katakana_ratio': round(katakana_ratio, 3),
            'ascii_ratio': round(ascii_ratio, 3),
            'other_ratio': round(other_ratio, 3),
            'style_type': self._guess_style(kanji_ratio),
        }

    def analyze_sentence_length(self, text: str) -> Dict:
        """
        文の長さの統計情報を計算

        Parameters:
            text (str): 入力テキスト

        Returns:
            Dict: 文長の統計
                {
                    'avg_length': float,
                    'std_length': float,
                    'min_length': int,
                    'max_length': int,
                    'sentence_count': int
                }
        """
        # 句読点で分割（。！？）
        sentences = re.split(r'[。！？\n]+', text)
        sentences = [s.strip() for s in sentences if s.strip()]

        if not sentences:
            return {
                'avg_length': 0.0,
                'std_length': 0.0,
                'min_length': 0,
                'max_length': 0,
                'sentence_count': 0,
            }

        lengths = [len(s) for s in sentences]

        return {
            'avg_length': round(np.mean(lengths), 1),
            'std_length': round(np.std(lengths), 1),
            'min_length': int(np.min(lengths)),
            'max_length': int(np.max(lengths)),
            'sentence_count': len(sentences),
        }

    def _assess_richness(self, yules_k: float) -> str:
        """
        Yule's K値から語彙の豊かさを評価

        Parameters:
            yules_k (float): Yule's K値

        Returns:
            str: 評価テキスト
        """
        if yules_k > 200:
            return "語彙の繰り返しが多い（単調）"
        elif yules_k > 150:
            return "やや単調"
        elif yules_k < 50:
            return "語彙が非常に多様（豊か）"
        elif yules_k < 100:
            return "語彙が多様"
        else:
            return "標準的な語彙多様性"

    def _guess_style(self, kanji_ratio: float) -> str:
        """
        漢字比率から文体を推定

        Parameters:
            kanji_ratio (float): 漢字の比率 (0.0-1.0)

        Returns:
            str: 推定される文体
        """
        if kanji_ratio > 0.45:
            return "硬質・論文調（学術的）"
        elif kanji_ratio > 0.35:
            return "標準的（バランス型）"
        elif kanji_ratio > 0.20:
            return "軟質・平易（親しみやすい）"
        else:
            return "非常に軟質（口語的）"

    # ─── Phase 2: 語彙多様性の深化 ─────────────────────────

    def analyze_advanced_diversity(self, text: str) -> Dict:
        """
        高度な語彙多様性指標群を計算

        TTR / Yule's K だけでは長文で崩壊する。
        以下の指標は長文耐性がある:

        - Hapax Legomena Ratio: 1回語の割合（語彙の洗練度）
        - Dis Legomena Ratio: 2回語の割合
        - Brunet's W: 対数ベース。長文に強い語彙多様性
        - Honoré's R: Hapax と総語数の統合指標
        - Sichel's S: Dis Legomena / Unique Types
        - Simpson's D: 「同じ語に2回当たる確率」の逆数

        Parameters:
            text: 入力テキスト

        Returns:
            Dict: 高度な語彙多様性指標
        """
        freq_dist = self._get_frequency_distribution(text)
        if not freq_dist:
            return self._empty_advanced_diversity()

        N = sum(freq_dist.values())       # 総トークン数
        V = len(freq_dist)                # ユニーク語彙数
        V1 = sum(1 for f in freq_dist.values() if f == 1)  # hapax legomena
        V2 = sum(1 for f in freq_dist.values() if f == 2)  # dis legomena

        # Hapax Legomena Ratio
        hapax_ratio = V1 / V if V > 0 else 0.0

        # Dis Legomena Ratio (Sichel's S)
        sichels_s = V2 / V if V > 0 else 0.0

        # Brunet's W: W = N^(V^-0.172)
        # 小さいほど語彙が豊か
        brunets_w = N ** (V ** -0.172) if V > 0 and N > 0 else 0.0

        # Honoré's R: R = 100 * log(N) / (1 - V1/V)
        # 大きいほど語彙が豊か
        if V > 0 and V1 < V:
            honores_r = 100.0 * math.log(N) / (1.0 - V1 / V)
        else:
            honores_r = 0.0

        # Simpson's Diversity Index: D = 1 - Σ(n_i * (n_i - 1)) / (N * (N - 1))
        # 1 に近いほど多様
        if N > 1:
            sum_ni_ni_1 = sum(f * (f - 1) for f in freq_dist.values())
            simpsons_d = 1.0 - sum_ni_ni_1 / (N * (N - 1))
        else:
            simpsons_d = 0.0

        return {
            'total_tokens': N,
            'unique_types': V,
            'hapax_legomena': V1,
            'dis_legomena': V2,
            'hapax_ratio': round(hapax_ratio, 4),
            'sichels_s': round(sichels_s, 4),
            'brunets_w': round(brunets_w, 4),
            'honores_r': round(honores_r, 2),
            'simpsons_d': round(simpsons_d, 6),
            'assessment': self._assess_advanced_diversity(
                hapax_ratio, brunets_w, simpsons_d
            ),
        }

    def analyze_mattr(self, text: str, window_size: int = 100) -> Dict:
        """
        MATTR (Moving Average Type-Token Ratio)

        固定ウィンドウをスライドさせて TTR を計算し、平均を取る。
        テキスト長に依存しない安定した語彙多様性指標。

        Parameters:
            text: 入力テキスト
            window_size: ウィンドウサイズ（トークン数）

        Returns:
            Dict: MATTR 結果と推移データ
        """
        tokens = self._tokenize_simple(text)
        n = len(tokens)

        if n < window_size:
            # ウィンドウより短い場合は通常の TTR
            unique = len(set(tokens))
            ttr = unique / n if n > 0 else 0.0
            return {
                'mattr': round(ttr, 4),
                'window_size': window_size,
                'num_windows': 1 if n > 0 else 0,
                'ttr_profile': [ttr] if n > 0 else [],
                'ttr_std': 0.0,
                'ttr_min': ttr,
                'ttr_max': ttr,
            }

        # スライディングウィンドウ
        ttr_values = []
        for i in range(n - window_size + 1):
            window = tokens[i:i + window_size]
            ttr = len(set(window)) / window_size
            ttr_values.append(ttr)

        mattr = float(np.mean(ttr_values))

        return {
            'mattr': round(mattr, 4),
            'window_size': window_size,
            'num_windows': len(ttr_values),
            'ttr_profile': [round(v, 4) for v in ttr_values],
            'ttr_std': round(float(np.std(ttr_values)), 4),
            'ttr_min': round(float(np.min(ttr_values)), 4),
            'ttr_max': round(float(np.max(ttr_values)), 4),
        }

    def analyze_vocabulary_growth(
        self, text: str, num_samples: int = 50
    ) -> Dict:
        """
        語彙成長曲線 (Vocabulary Growth Curve) と Heaps' Law フィッティング

        Heaps' Law: V(n) = K * n^β
        - K: 定数（高いほど初期語彙が豊富）
        - β: 成長率（1に近いほど新語が出続ける。0.4-0.6が典型）

        Parameters:
            text: 入力テキスト
            num_samples: サンプリングポイント数

        Returns:
            Dict: 成長曲線データとフィッティングパラメータ
        """
        tokens = self._tokenize_simple(text)
        n = len(tokens)

        if n < 10:
            return {
                'heaps_K': 0.0,
                'heaps_beta': 0.0,
                'growth_curve': [],
                'assessment': '分析対象が少なすぎます',
            }

        # サンプリングポイントでの語彙数を計測
        sample_points = np.linspace(10, n, min(num_samples, n // 2), dtype=int)
        sample_points = sorted(set(sample_points))

        growth = []
        seen = set()
        token_idx = 0

        for target_n in sample_points:
            while token_idx < target_n and token_idx < n:
                seen.add(tokens[token_idx])
                token_idx += 1
            growth.append({'n': int(target_n), 'V': len(seen)})

        # Heaps' Law フィッティング: log(V) = log(K) + β * log(n)
        if len(growth) >= 3:
            log_n = np.array([math.log(g['n']) for g in growth])
            log_v = np.array([math.log(g['V']) for g in growth if g['V'] > 0])

            if len(log_n) == len(log_v) and len(log_v) >= 2:
                # 最小二乗法
                coeffs = np.polyfit(log_n[:len(log_v)], log_v, 1)
                beta = float(coeffs[0])
                K = float(math.exp(coeffs[1]))
            else:
                beta = 0.0
                K = 0.0
        else:
            beta = 0.0
            K = 0.0

        return {
            'heaps_K': round(K, 4),
            'heaps_beta': round(beta, 4),
            'growth_curve': growth,
            'total_tokens': n,
            'final_vocabulary': len(seen),
            'assessment': self._assess_heaps(beta),
        }

    # ─── Private Helpers ────────────────────────────────────

    def _get_frequency_distribution(self, text: str) -> Counter:
        """テキストから簡易的な語彙頻度分布を取得"""
        tokens = self._tokenize_simple(text)
        return Counter(tokens)

    def _tokenize_simple(self, text: str) -> List[str]:
        """
        簡易トークナイザ（2-4文字 N-gram ベース）

        形態素解析不要。文字 N-gram で語彙多様性を近似。
        """
        clean = re.sub(r'\s+', '', text)
        if len(clean) < 2:
            return list(clean)

        tokens = []
        # 2文字と3文字のN-gramを語彙単位として使用
        for n in [2, 3]:
            for i in range(len(clean) - n + 1):
                gram = clean[i:i + n]
                tokens.append(gram)
        return tokens

    @staticmethod
    def _empty_advanced_diversity() -> Dict:
        return {
            'total_tokens': 0, 'unique_types': 0,
            'hapax_legomena': 0, 'dis_legomena': 0,
            'hapax_ratio': 0.0, 'sichels_s': 0.0,
            'brunets_w': 0.0, 'honores_r': 0.0,
            'simpsons_d': 0.0, 'assessment': '分析対象がありません',
        }

    @staticmethod
    def _assess_advanced_diversity(
        hapax_ratio: float, brunets_w: float, simpsons_d: float
    ) -> str:
        """高度な多様性指標から総合評価"""
        scores = 0
        if hapax_ratio > 0.5:
            scores += 2  # 半数以上が1回語 = 非常に豊か
        elif hapax_ratio > 0.3:
            scores += 1

        if simpsons_d > 0.98:
            scores += 2
        elif simpsons_d > 0.95:
            scores += 1

        if scores >= 3:
            return "語彙が非常に豊か（高度に多様）"
        elif scores >= 2:
            return "語彙が豊か"
        elif scores >= 1:
            return "標準的な語彙多様性"
        else:
            return "語彙がやや単調"

    @staticmethod
    def _assess_heaps(beta: float) -> str:
        """Heaps' β から語彙成長を評価"""
        if beta <= 0:
            return "フィッティング不可"
        elif beta > 0.7:
            return "新語が非常に多い（語彙が拡散的）"
        elif beta > 0.5:
            return "新語の出現が活発"
        elif beta > 0.3:
            return "標準的な語彙成長"
        else:
            return "語彙が収束している（繰り返しが多い）"

    # ─── analyze_full 拡張 ──────────────────────────────────

    def analyze_full(self, text: str) -> Dict:
        """
        テキストの全体的な文体分析を実施（強化版）

        Parameters:
            text (str): 入力テキスト

        Returns:
            Dict: 包括的な文体分析結果
        """
        return {
            'vocabulary': self.analyze_vocabulary_richness(text),
            'advanced_diversity': self.analyze_advanced_diversity(text),
            'mattr': self.analyze_mattr(text),
            'vocabulary_growth': self.analyze_vocabulary_growth(text),
            'character_type': self.analyze_char_type_ratio(text),
            'sentence': self.analyze_sentence_length(text),
        }


# 便利関数
def analyze_stylometry(text: str) -> Dict:
    """
    単一関数での全体的な文体分析

    Parameters:
        text (str): 入力テキスト

    Returns:
        Dict: 全体的な文体分析結果
    """
    analyzer = StylometryAnalyzer()
    return analyzer.analyze_full(text)


def get_stylometry_summary(text: str) -> str:
    """
    文体分析の簡潔なサマリーを返す（強化版）

    Parameters:
        text (str): 入力テキスト

    Returns:
        str: フォーマットされたサマリー
    """
    analyzer = StylometryAnalyzer()
    result = analyzer.analyze_full(text)

    summary = []
    summary.append("=" * 60)
    summary.append("【文体分析結果（強化版）】")
    summary.append("=" * 60)

    # 基本語彙多様性
    vocab = result.get('vocabulary', {})
    summary.append("\n【基本語彙多様性】")
    summary.append(f"  TTR: {vocab.get('ttr', 0):.4f}")
    summary.append(f"  Yule's K: {vocab.get('yules_k', 0)}")
    summary.append(f"  評価: {vocab.get('assessment', 'N/A')}")

    # 高度な語彙多様性
    adv = result.get('advanced_diversity', {})
    summary.append("\n【高度な語彙多様性】")
    summary.append(f"  Hapax Legomena: {adv.get('hapax_legomena', 0)} ({adv.get('hapax_ratio', 0):.1%})")
    summary.append(f"  Sichel's S:    {adv.get('sichels_s', 0):.4f}")
    summary.append(f"  Brunet's W:    {adv.get('brunets_w', 0):.4f}")
    summary.append(f"  Honoré's R:    {adv.get('honores_r', 0):.2f}")
    summary.append(f"  Simpson's D:   {adv.get('simpsons_d', 0):.6f}")
    summary.append(f"  評価: {adv.get('assessment', 'N/A')}")

    # MATTR
    mattr = result.get('mattr', {})
    summary.append("\n【MATTR (Moving Average TTR)】")
    summary.append(f"  MATTR: {mattr.get('mattr', 0):.4f}")
    summary.append(f"  TTR変動(σ): {mattr.get('ttr_std', 0):.4f}")
    summary.append(f"  TTR範囲: {mattr.get('ttr_min', 0):.4f} - {mattr.get('ttr_max', 0):.4f}")

    # 語彙成長
    growth = result.get('vocabulary_growth', {})
    summary.append("\n【語彙成長曲線 (Heaps' Law)】")
    summary.append(f"  K (初期語彙量): {growth.get('heaps_K', 0):.4f}")
    summary.append(f"  β (成長率):     {growth.get('heaps_beta', 0):.4f}")
    summary.append(f"  評価: {growth.get('assessment', 'N/A')}")

    # 文字種比率
    char_type = result.get('character_type', {})
    summary.append("\n【文字種比率】")
    summary.append(f"  漢字: {char_type.get('kanji_ratio', 0):.1%}")
    summary.append(f"  ひらがな: {char_type.get('hiragana_ratio', 0):.1%}")
    summary.append(f"  カタカナ: {char_type.get('katakana_ratio', 0):.1%}")
    summary.append(f"  文体: {char_type.get('style_type', 'N/A')}")

    # 文長統計
    sent = result.get('sentence', {})
    summary.append("\n【文長統計】")
    summary.append(f"  平均文長: {sent.get('avg_length', 0):.1f}字")
    summary.append(f"  標準偏差: {sent.get('std_length', 0):.1f}字")
    summary.append(f"  文の数: {sent.get('sentence_count', 0)}")

    summary.append("\n" + "=" * 60)
    return "\n".join(summary)



# coding: utf-8
"""
重複検出・削除ロジック

JSON出力やスコア候補の重複を統一的に検出・削除
"""

import logging
from typing import List, Dict, Any, Set, Tuple
from collections import defaultdict

logger = logging.getLogger(__name__)


class DuplicateHandler:
    """重複検出・削除の共通ロジック"""

    @staticmethod
    def deduplicate_candidates(candidates: List[Dict[str, Any]], 
                              key_func=None,
                              merge_func=None) -> List[Dict[str, Any]]:
        """
        候補リストから重複を削除

        Args:
            candidates: 候補リスト
            key_func: 重複判定のキーを生成する関数
            merge_func: 重複した候補をマージする関数

        Returns:
            重複排除後の候補リスト
        """
        if not candidates:
            return []

        if key_func is None:
            # デフォルト: 'primary' フィールドで判定
            key_func = lambda c: c.get('primary', '')

        if merge_func is None:
            # デフォルト: 最初の候補を保持、confidence が高い方を優先
            merge_func = lambda c1, c2: c1 if c1.get('confidence', 0) >= c2.get('confidence', 0) else c2

        seen = {}
        duplicates = defaultdict(list)

        for candidate in candidates:
            key = key_func(candidate)
            if key in seen:
                duplicates[key].append(candidate)
            else:
                seen[key] = candidate

        # 重複候補をマージ
        for key, dup_list in duplicates.items():
            original = seen[key]
            for dup in dup_list:
                merged = merge_func(original, dup)
                seen[key] = merged

        result = list(seen.values())
        if len(result) < len(candidates):
            logger.info(f"重複検出: {len(candidates)} → {len(result)} ({len(candidates) - len(result)} 件削除)")

        return result

    @staticmethod
    def find_duplicate_positions(issues: List[Dict[str, Any]]) -> List[List[int]]:
        """
        位置情報が重複している問題を検出

        Args:
            issues: 問題リスト（'position' フィールド必須）

        Returns:
            重複グループのインデックスリスト
        """
        if not issues:
            return []

        position_map = defaultdict(list)
        for i, issue in enumerate(issues):
            if 'position' in issue:
                pos = issue['position']
                # (start, end) をキーに
                key = (pos[0], pos[1])
                position_map[key].append(i)

        # 同じ位置の問題グループを抽出
        duplicates = [indices for indices in position_map.values() if len(indices) > 1]
        
        if duplicates:
            logger.debug(f"位置重複検出: {len(duplicates)} グループ")

        return duplicates

    @staticmethod
    def merge_issues_at_position(issues: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        同じ位置の問題をマージ

        Args:
            issues: 問題リスト

        Returns:
            マージ後の問題リスト
        """
        duplicates = DuplicateHandler.find_duplicate_positions(issues)

        if not duplicates:
            return issues

        # 重複する位置のインデックスセット
        dup_indices = set()
        for group in duplicates:
            dup_indices.update(group)

        # マージ対象を整理
        merged_issues = []
        processed = set()

        for i, issue in enumerate(issues):
            if i in processed:
                continue

            if i in dup_indices:
                # 同じ位置のすべての問題を集計
                pos = issue['position']
                same_pos_issues = [issues[j] for j in dup_indices if issues[j]['position'] == pos]

                # マージ問題を作成
                merged = {
                    'position': pos,
                    'type': 'merged_issues',
                    'count': len(same_pos_issues),
                    'original_types': list(set(i.get('type', 'unknown') for i in same_pos_issues)),
                    'original_issues': same_pos_issues,
                    'messages': [i.get('message', '') for i in same_pos_issues if i.get('message')]
                }

                merged_issues.append(merged)
                processed.update([j for j, iss in enumerate(issues) if iss['position'] == pos])
            else:
                merged_issues.append(issue)
                processed.add(i)

        if len(merged_issues) < len(issues):
            logger.info(f"位置重複マージ: {len(issues)} → {len(merged_issues)} ({len(issues) - len(merged_issues)} 件統合)")

        return merged_issues

    @staticmethod
    def filter_duplicate_suggestions(suggestions: List[Dict[str, Any]],
                                     position_tolerance: int = 0) -> List[Dict[str, Any]]:
        """
        修正提案から近い位置の重複を削除

        Args:
            suggestions: 修正提案リスト（'position' フィールド必須）
            position_tolerance: 同一視する位置の許容範囲（文字数）

        Returns:
            重複排除後の提案リスト
        """
        if not suggestions:
            return []

        # 位置でグループ化
        position_groups = defaultdict(list)
        for sugg in suggestions:
            if 'position' not in sugg:
                continue

            # 許容範囲内の位置をキーにグループ化
            pos = sugg['position']
            key = (
                pos // (position_tolerance + 1) if position_tolerance > 0 else pos
            )
            position_groups[key].append(sugg)

        # グループごとに信頼度が最も高いものを保持
        result = []
        for key, group in position_groups.items():
            if len(group) > 1:
                logger.debug(f"位置 {key} で {len(group)} 件の提案が重複")
                # confidence が最も高いものを選択
                best = max(group, 
                          key=lambda s: s.get('confidence', 0))
                result.append(best)
            else:
                result.append(group[0])

        if len(result) < len(suggestions):
            logger.info(f"提案重複排除: {len(suggestions)} → {len(result)} ({len(suggestions) - len(result)} 件削除)")

        return result

    @staticmethod
    def detect_score_conflicts(candidates: List[Dict[str, Any]]) -> List[Tuple[int, int, float]]:
        """
        スコアに矛盾がある候補対を検出

        Args:
            candidates: 候補リスト（'confidence', 'primary', 'variants' フィールド必須）

        Returns:
            (index1, index2, conflict_score) のリスト
        """
        conflicts = []

        for i, cand1 in enumerate(candidates):
            for j, cand2 in enumerate(candidates[i+1:], i+1):
                # 基準フレーズが重複している場合
                if cand1.get('primary') == cand2.get('primary'):
                    conf1 = cand1.get('confidence', 0)
                    conf2 = cand2.get('confidence', 0)
                    conflict_score = abs(conf1 - conf2)

                    if conflict_score > 0.1:  # 信頼度の差が10%以上
                        conflicts.append((i, j, conflict_score))
                        logger.warning(
                            f"スコア矛盾検出: {cand1.get('primary')} - "
                            f"conf1={conf1:.2f}, conf2={conf2:.2f}"
                        )

        return conflicts



"""
統計的スコアリング（PMI・分岐エントロピー）のテスト
"""

import pytest
import numpy as np
import pandas as pd
from japhrase import PhraseExtracter


class TestPMICalculation:
    """PMI（自己相互情報量）計算のテスト"""

    def test_calculate_pmi_high_cohesion(self):
        """結合度が高いフレーズのPMI"""
        extractor = PhraseExtracter(use_pmi=True)
        phrases = ["機械学習"]
        text = "機械学習は機械学習で機械学習を"  # "機械学習"が繰り返される

        pmi_scores = extractor.calculate_pmi(phrases, text)

        # 結合度が高いので、PMIは正の値
        assert "機械学習" in pmi_scores
        assert pmi_scores["機械学習"] > 0

    def test_calculate_pmi_low_cohesion(self):
        """結合度が低いフレーズのPMI"""
        extractor = PhraseExtracter(use_pmi=True)
        # 「ab」は「a」と「b」の独立確率の積に近い → PMIが低い
        phrases = ["ab"]
        # 「a」「b」が個別に頻出するがペアでは少ない
        text = "aaaabbbbab"  # 「ab」は1回、「a」は5回、「b」は5回

        pmi_scores = extractor.calculate_pmi(phrases, text)

        assert "ab" in pmi_scores
        # PMIスコアが計算される
        assert isinstance(pmi_scores["ab"], float)

    def test_calculate_pmi_single_char(self):
        """単一文字のPMI"""
        extractor = PhraseExtracter(use_pmi=True)
        phrases = ["あ"]
        text = "あああああああ"

        pmi_scores = extractor.calculate_pmi(phrases, text)

        assert "あ" in pmi_scores
        # 単一文字なのでPMIは0に近い
        assert pmi_scores["あ"] >= 0

    def test_calculate_pmi_not_found(self):
        """フレーズが見つからない場合"""
        extractor = PhraseExtracter(use_pmi=True)
        phrases = ["存在しないフレーズ"]
        text = "これはテストです"

        pmi_scores = extractor.calculate_pmi(phrases, text)

        assert "存在しないフレーズ" in pmi_scores
        assert pmi_scores["存在しないフレーズ"] == 0.0

    def test_calculate_pmi_empty_input(self):
        """空の入力"""
        extractor = PhraseExtracter(use_pmi=True)
        pmi_scores = extractor.calculate_pmi([], "")

        assert pmi_scores == {}

    def test_pmi_numerical_stability(self):
        """PMIの数値安定性（クリップ処理）"""
        extractor = PhraseExtracter(use_pmi=True)
        phrases = ["x"]
        text = "x" * 1000  # 非常に高いPMI

        pmi_scores = extractor.calculate_pmi(phrases, text)

        # PMIが数値的に安定している（-10 ~ 10の範囲）
        assert -10 <= pmi_scores["x"] <= 10


class TestBranchingEntropyCalculation:
    """分岐エントロピー（Branching Entropy）計算のテスト"""

    def test_branching_entropy_word_boundary(self):
        """単語境界でのエントロピー"""
        extractor = PhraseExtracter(use_branching_entropy=True)
        sentences = ["機械学習は重要です。機械学習を学ぶ。"]
        phrases = ["機械学習"]

        entropy_scores = extractor.calculate_branching_entropy(sentences, phrases)

        assert "機械学習" in entropy_scores
        left_entropy, right_entropy, boundary_score = entropy_scores["機械学習"]

        # 単語境界なので、左右のエントロピーが比較的高い
        assert left_entropy >= 0
        assert right_entropy >= 0
        assert boundary_score >= 0

    def test_branching_entropy_word_middle(self):
        """単語の途中でのエントロピー"""
        extractor = PhraseExtracter(use_branching_entropy=True)
        sentences = ["テストテストテスト"]  # "テス"の後は常に"ト"
        phrases = ["テス"]

        entropy_scores = extractor.calculate_branching_entropy(sentences, phrases)

        assert "テス" in entropy_scores
        left_entropy, right_entropy, boundary_score = entropy_scores["テス"]

        # 単語の途中なので、右エントロピーが低い（「ト」しか来ない）
        assert right_entropy < 2.0  # 低いエントロピー

    def test_branching_entropy_single_occurrence(self):
        """フレーズが1回だけ出現"""
        extractor = PhraseExtracter(use_branching_entropy=True)
        sentences = ["テストのみ"]
        phrases = ["テスト"]

        entropy_scores = extractor.calculate_branching_entropy(sentences, phrases)

        assert "テスト" in entropy_scores
        left_entropy, right_entropy, boundary_score = entropy_scores["テスト"]

        # 1回だけの出現なので、エントロピーは0か非常に小さい
        assert boundary_score >= 0
        assert boundary_score <= 1.0

    def test_branching_entropy_empty_input(self):
        """空の入力"""
        extractor = PhraseExtracter(use_branching_entropy=True)
        entropy_scores = extractor.calculate_branching_entropy([], [])

        assert entropy_scores == {}

    def test_branching_entropy_normalization(self):
        """エントロピースコアが0-1に正規化されている"""
        extractor = PhraseExtracter(use_branching_entropy=True)
        sentences = ["あああああいいいううう"]
        phrases = ["あ", "い", "う"]

        entropy_scores = extractor.calculate_branching_entropy(sentences, phrases)

        for phrase in phrases:
            left_entropy, right_entropy, boundary_score = entropy_scores[phrase]
            # スコアは0-1の範囲
            assert 0.0 <= boundary_score <= 1.0


class TestScoringIntegration:
    """PMI・BEのスコアリング統合テスト"""

    def test_score_with_pmi_enabled(self):
        """PMI有効時のスコアリング"""
        extractor_with_pmi = PhraseExtracter(
            min_count=2,
            use_pmi=True,
            verbose=0
        )
        extractor_without_pmi = PhraseExtracter(
            min_count=2,
            use_pmi=False,
            verbose=0
        )

        sentences = ["機械学習は機械学習で機械学習", "ていうのはていう"]

        # PMI有効版と無効版でスコアが異なる
        df_with = extractor_with_pmi.get_dfphrase(sentences)
        df_without = extractor_without_pmi.get_dfphrase(sentences)

        # 両方とも結果を返す（エラーがない）
        assert isinstance(df_with, pd.DataFrame)
        assert isinstance(df_without, pd.DataFrame)

    def test_score_with_entropy_enabled(self):
        """BE有効時のスコアリング"""
        extractor_with_entropy = PhraseExtracter(
            min_count=2,
            use_branching_entropy=True,
            verbose=0
        )
        extractor_without_entropy = PhraseExtracter(
            min_count=2,
            use_branching_entropy=False,
            verbose=0
        )

        sentences = ["機械学習は機械学習で機械学習です", "テストテストテスト"]

        df_with = extractor_with_entropy.get_dfphrase(sentences)
        df_without = extractor_without_entropy.get_dfphrase(sentences)

        # 両方とも結果を返す（エラーがない）
        assert isinstance(df_with, pd.DataFrame)
        assert isinstance(df_without, pd.DataFrame)

    def test_score_with_both_enabled(self):
        """PMI＋BE両方有効時のスコアリング"""
        extractor = PhraseExtracter(
            min_count=2,
            use_pmi=True,
            use_branching_entropy=True,
            pmi_weight=1.0,
            entropy_weight=1.0,
            verbose=0
        )

        sentences = [
            "機械学習は重要です。機械学習を学びます。",
            "ていうのはていう意味で",
            "テストテストテスト",
        ]

        df = extractor.get_dfphrase(sentences)

        # 結果が返される（エラーがない）
        assert isinstance(df, pd.DataFrame)
        # フレーズが抽出されている
        if len(df) > 0:
            assert "seqchar" in df.columns
            assert "freq" in df.columns
            assert "sc_index" in df.columns

    def test_backward_compatibility(self):
        """後方互換性（デフォルトではPMI/BEを使用しない）"""
        extractor_default = PhraseExtracter(verbose=0)
        assert extractor_default.use_pmi == False
        assert extractor_default.use_branching_entropy == False

        sentences = ["テストテストテスト", "これはテストです"]
        df = extractor_default.get_dfphrase(sentences)

        # PMI/BE列がない（デフォルト動作）
        assert "pmi" not in df.columns
        assert "boundary_score" not in df.columns


class TestParameterWeights:
    """PMI・BEの重み係数テスト"""

    def test_pmi_weight_parameter(self):
        """PMI重み係数の影響"""
        sentences = ["機械学習は機械学習で機械学習"]

        # 異なる重みでスコアが変わることを確認
        extractor1 = PhraseExtracter(
            min_count=2,
            use_pmi=True,
            pmi_weight=0.5,
            verbose=0
        )
        extractor2 = PhraseExtracter(
            min_count=2,
            use_pmi=True,
            pmi_weight=2.0,
            verbose=0
        )

        df1 = extractor1.get_dfphrase(sentences)
        df2 = extractor2.get_dfphrase(sentences)

        # 両方とも結果を返す
        assert isinstance(df1, pd.DataFrame)
        assert isinstance(df2, pd.DataFrame)

    def test_entropy_weight_parameter(self):
        """エントロピー重み係数の影響"""
        sentences = ["機械学習は機械学習で機械学習です"]

        # 異なる重みでスコアが変わることを確認
        extractor1 = PhraseExtracter(
            min_count=2,
            use_branching_entropy=True,
            entropy_weight=0.5,
            verbose=0
        )
        extractor2 = PhraseExtracter(
            min_count=2,
            use_branching_entropy=True,
            entropy_weight=2.0,
            verbose=0
        )

        df1 = extractor1.get_dfphrase(sentences)
        df2 = extractor2.get_dfphrase(sentences)

        # 両方とも結果を返す
        assert isinstance(df1, pd.DataFrame)
        assert isinstance(df2, pd.DataFrame)


class TestEdgeCases:
    """エッジケースのテスト"""

    def test_empty_phrase_list(self):
        """フレーズが空の場合"""
        extractor = PhraseExtracter(use_pmi=True, use_branching_entropy=True)

        pmi_scores = extractor.calculate_pmi([], "テキスト")
        entropy_scores = extractor.calculate_branching_entropy(["テキスト"], [])

        assert pmi_scores == {}
        assert entropy_scores == {}

    def test_very_long_text(self):
        """非常に長いテキスト"""
        extractor = PhraseExtracter(
            min_count=2,
            use_pmi=True,
            use_branching_entropy=True,
            verbose=0
        )

        # 長いテキストを生成
        long_text = "機械学習 " * 100
        sentences = [long_text]

        df = extractor.get_dfphrase(sentences)

        # エラーなく処理できる
        assert isinstance(df, pd.DataFrame)

    def test_special_characters(self):
        """特殊文字を含むテキスト"""
        extractor = PhraseExtracter(
            min_count=2,
            use_pmi=True,
            use_branching_entropy=True,
            verbose=0
        )

        sentences = ["テスト★テスト★テスト", "ABC★ABC★ABC"]

        df = extractor.get_dfphrase(sentences)

        # エラーなく処理できる
        assert isinstance(df, pd.DataFrame)

    def test_unicode_characters(self):
        """Unicode文字を含むテキスト"""
        extractor = PhraseExtracter(
            min_count=2,
            use_pmi=True,
            use_branching_entropy=True,
            verbose=0
        )

        sentences = ["😀😀😀", "🎉🎉🎉", "機械学習は機械学習"]

        df = extractor.get_dfphrase(sentences)

        # エラーなく処理できる
        assert isinstance(df, pd.DataFrame)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# coding: utf-8
"""
フレーズ統一エンジン

検出された表記ゆれを実際に統一する機能
"""

import logging
from typing import List, Dict, Tuple, Optional
import pandas as pd
import re

logger = logging.getLogger(__name__)


class PhraseUnifier:
    """フレーズの統一を実行"""

    def __init__(self):
        self.unification_map = {}  # variant → primary のマッピング
        self.unification_history = []

    def unify_phrases(
        self,
        df_phrases: pd.DataFrame,
        primary: str,
        variants: List[str],
        apply_to_texts: bool = False,
        texts: Optional[List[str]] = None
    ) -> Tuple[pd.DataFrame, Optional[List[str]]]:
        """
        表記ゆれを統一

        Parameters:
            df_phrases: フレーズDataFrame
            primary: 基準フレーズ
            variants: 統一対象の候補リスト
            apply_to_texts: 元テキストにも適用するか
            texts: 元テキスト（apply_to_texts=True の場合）

        Returns:
            (統一後のDataFrame, 統一後のテキスト)
        """
        if 'seqchar' not in df_phrases.columns and 'phrase' not in df_phrases.columns:
            logger.error("フレーズ列が見つかりません")
            return df_phrases, texts

        phrase_col = 'seqchar' if 'seqchar' in df_phrases.columns else 'phrase'

        # 統一前の統計
        before_count = len(df_phrases[df_phrases[phrase_col] == primary])
        before_freq = df_phrases[df_phrases[phrase_col] == primary]['freq'].sum() if 'freq' in df_phrases.columns else before_count

        # variant フレーズの出現回数を集計
        variant_freq = df_phrases[df_phrases[phrase_col].isin(variants)]['freq'].sum() \
                      if 'freq' in df_phrases.columns else len(df_phrases[df_phrases[phrase_col].isin(variants)])

        # マッピングテーブルを更新
        for variant in variants:
            self.unification_map[variant] = primary

        # DataFrameを複製して更新
        df_unified = df_phrases.copy()

        # variant をすべて primary に置換
        for variant in variants:
            mask = df_unified[phrase_col] == variant
            df_unified.loc[mask, phrase_col] = primary

        # 同じフレーズが複数行ある場合は集約
        if 'freq' in df_unified.columns:
            df_unified = df_unified.groupby(phrase_col, as_index=False).agg({
                'freq': 'sum',
                **{col: 'first' for col in df_unified.columns if col not in [phrase_col, 'freq']}
            })

        # 統一後の統計
        after_count = len(df_unified[df_unified[phrase_col] == primary])
        after_freq = df_unified[df_unified[phrase_col] == primary]['freq'].sum() if 'freq' in df_unified.columns else after_count

        # 統一履歴に記録
        self.unification_history.append({
            'primary': primary,
            'variants': variants,
            'before_freq': before_freq,
            'after_freq': after_freq,
            'freq_gain': after_freq - before_freq,
            'variant_count': len(variants),
        })

        # テキストも統一
        texts_unified = None
        if apply_to_texts and texts:
            texts_unified = self._unify_texts(texts, primary, variants)

        logger.info(
            f"✓ 統一完了: '{primary}' ({before_freq} → {after_freq} 回, +{after_freq - before_freq})"
        )

        return df_unified, texts_unified

    def _unify_texts(self, texts: List[str], primary: str, variants: List[str]) -> List[str]:
        """元テキストのフレーズを統一"""
        texts_unified = []

        for text in texts:
            unified_text = text
            for variant in variants:
                # 単語境界を考慮して置換（オプション）
                unified_text = unified_text.replace(variant, primary)

            texts_unified.append(unified_text)

        logger.info(f"テキスト内のフレーズを統一しました ({len(texts)} 行)")
        return texts_unified

    def batch_unify(
        self,
        df_phrases: pd.DataFrame,
        unification_rules: List[Dict],
        texts: Optional[List[str]] = None
    ) -> Tuple[pd.DataFrame, Optional[List[str]]]:
        """
        複数の統一ルールを一括適用

        Parameters:
            df_phrases: フレーズDataFrame
            unification_rules: 統一ルールのリスト
                [
                    {'primary': 'テスト', 'variants': ['テスツ', 'てすと']},
                    {'primary': '実施', 'variants': ['実行']},
                ]
            texts: 元テキスト

        Returns:
            (統一後のDataFrame, 統一後のテキスト)
        """
        df_result = df_phrases.copy()
        texts_result = texts.copy() if texts else None

        for rule in unification_rules:
            primary = rule['primary']
            variants = rule['variants']

            df_result, texts_result = self.unify_phrases(
                df_result,
                primary=primary,
                variants=variants,
                apply_to_texts=(texts_result is not None),
                texts=texts_result
            )

        logger.info(f"✓ {len(unification_rules)} 件の統一ルールを適用しました")
        return df_result, texts_result

    def get_unification_stats(self) -> Dict:
        """統一処理の統計を取得"""
        if not self.unification_history:
            return {}

        total_freq_gain = sum(h['freq_gain'] for h in self.unification_history)
        total_variants = sum(h['variant_count'] for h in self.unification_history)

        return {
            'unification_count': len(self.unification_history),
            'total_variants': total_variants,
            'total_freq_gain': total_freq_gain,
            'history': self.unification_history,
        }

    def export_stats(self, filepath: str) -> None:
        """統一処理の統計を CSV に出力"""
        df_stats = pd.DataFrame(self.unification_history)
        df_stats.to_csv(filepath, index=False, encoding='utf-8-sig')
        logger.info(f"統一統計を {filepath} に出力しました")

    def generate_report(self) -> str:
        """統一処理のレポートを生成"""
        if not self.unification_history:
            return "統一処理の履歴がありません"

        report = "【フレーズ統一レポート】\n\n"

        total_freq_gain = sum(h['freq_gain'] for h in self.unification_history)
        total_variants = sum(h['variant_count'] for h in self.unification_history)

        report += f"統一処理数: {len(self.unification_history)}\n"
        report += f"統一されたバリアント総数: {total_variants}\n"
        report += f"出現回数の増加: +{total_freq_gain}\n\n"

        report += "【詳細】\n"
        for i, h in enumerate(self.unification_history, 1):
            report += (
                f"{i}. '{h['primary']}' "
                f"(バリアント: {h['variant_count']}個, "
                f"出現回数: {h['before_freq']} → {h['after_freq']} +{h['freq_gain']})\n"
            )
            report += f"   統一内容: {', '.join(self.unification_map.get(h['primary'], []))}\n"

        return report


class InteractiveUnifier:
    """対話的に表記ゆれを統一"""

    def __init__(self, detector, unifier):
        self.detector = detector
        self.unifier = unifier

    def interactive_unify(
        self,
        df_phrases: pd.DataFrame,
        texts: Optional[List[str]] = None,
        top_k: int = 20
    ) -> Tuple[pd.DataFrame, Optional[List[str]]]:
        """
        対話的にユーザーに候補を提示して統一

        Parameters:
            df_phrases: フレーズDataFrame
            texts: 元テキスト（コンテキスト分析用）
            top_k: 提示する候補の上限数

        Returns:
            (統一後のDataFrame, 統一後のテキスト)
        """
        # 表記ゆれ候補を検出
        candidates = self.detector.detect_variants(df_phrases, texts)

        if not candidates:
            print("表記ゆれ候補が見つかりませんでした")
            return df_phrases, texts

        df_result = df_phrases.copy()
        texts_result = texts.copy() if texts else None

        print(f"\n表記ゆれ候補 {len(candidates[:top_k])} 件が見つかりました\n")

        accepted_count = 0
        rejected_count = 0

        for i, candidate in enumerate(candidates[:top_k], 1):
            print(f"\n{'=' * 80}")
            print(f"【候補 {i}/{min(top_k, len(candidates))}】\n")

            self._display_candidate(candidate, i)

            # ユーザー入力
            while True:
                choice = input(
                    f"\nこの統一を実行? (y=yes, n=skip, q=quit): "
                ).strip().lower()

                if choice == 'y':
                    # 統一を実行
                    df_result, texts_result = self.unifier.unify_phrases(
                        df_result,
                        primary=candidate.primary,
                        variants=candidate.variants,
                        apply_to_texts=(texts_result is not None),
                        texts=texts_result
                    )
                    accepted_count += 1
                    break

                elif choice == 'n':
                    print("  → スキップしました")
                    rejected_count += 1
                    break

                elif choice == 'q':
                    print("\n対話を終了します")
                    return df_result, texts_result

                else:
                    print("  入力が無効です。y/n/q で入力してください")

        print(f"\n{'=' * 80}")
        print(f"\n【統一完了】")
        print(f"  受け入れ: {accepted_count} 件")
        print(f"  スキップ: {rejected_count} 件")
        print(f"  フレーズ数の変化: {len(df_phrases)} → {len(df_result)}")

        return df_result, texts_result

    def _display_candidate(self, candidate, index: int) -> None:
        """候補を見やすく表示"""
        print(f"基準フレーズ: '{candidate.primary}' (出現: {candidate.primary_freq}回)")
        print(f"表記ゆれ候補:")
        for variant, freq in zip(candidate.variants, candidate.variant_freqs):
            print(f"  • {variant} (出現: {freq}回)")

        print(f"\n【統計的根拠】")
        print(f"  編集距離スコア: {candidate.edit_distance_score:.2f}")
        print(f"    → 文字レベルの距離が近い")

        print(f"  文字種変化: {candidate.char_type_shift_score:.2f}")
        if candidate.char_type_changes:
            print(f"    → {', '.join(candidate.char_type_changes)}")

        print(f"  長さの比率: {candidate.length_ratio_score:.2f}")
        print(f"  共通接頭辞: {candidate.common_prefix_score:.2f}")

        print(f"  コンテキスト類似度: {candidate.context_similarity_score:.2f}")
        print(f"    → 前後の単語が {candidate.context_similarity_score:.0%} 重複")

        print(f"  同時出現率: {candidate.co_occurrence_score:.2f}")
        if candidate.co_occur_count == 0:
            print(f"    → 同じテキストでの同時出現: なし")
        else:
            print(f"    → 同じテキストでの同時出現: {candidate.co_occur_count}回")

        print(f"\n【評価】")
        print(f"  確信度: {candidate.confidence:.1%}")
        print(f"  推奨: {candidate.recommendation}")
        print(f"  理由: {candidate.reasoning}")

        freq_gain = sum(candidate.variant_freqs)
        print(f"\n  効果: '{candidate.primary}' の頻出度が +{freq_gain}回 向上")



# coding: utf-8
"""
Generate a minimal HTML report from stats data.
"""

from __future__ import annotations

from html import escape
from typing import Dict, List


def _svg_bar_chart(
    items: List[Dict],
    label_key: str,
    value_key: str,
    width: int = 720,
    height: int = 320,
    bar_color: str = "#2a9d8f",
) -> str:
    if not items:
        return "<svg width=\"{0}\" height=\"{1}\"></svg>".format(width, height)

    max_value = max(item[value_key] for item in items) or 1
    padding = 40
    chart_width = width - padding * 2
    chart_height = height - padding * 2
    bar_width = chart_width / max(len(items), 1)

    bars = []
    labels = []
    for idx, item in enumerate(items):
        value = item[value_key]
        bar_height = (value / max_value) * chart_height
        x = padding + idx * bar_width
        y = padding + (chart_height - bar_height)
        bars.append(
            f'<rect x="{x:.1f}" y="{y:.1f}" width="{bar_width - 4:.1f}" '
            f'height="{bar_height:.1f}" fill="{bar_color}"></rect>'
        )
        label = escape(str(item[label_key]))
        labels.append(
            f'<text x="{x + 2:.1f}" y="{height - 8}" font-size="10" '
            f'fill="#333" transform="rotate(45 {x + 2:.1f},{height - 8})">{label}</text>'
        )

    axis = (
        f'<line x1="{padding}" y1="{padding}" x2="{padding}" y2="{height - padding}" '
        f'stroke="#444" stroke-width="1" />'
        f'<line x1="{padding}" y1="{height - padding}" x2="{width - padding}" '
        f'y2="{height - padding}" stroke="#444" stroke-width="1" />'
    )

    return (
        f'<svg width="{width}" height="{height}" viewBox="0 0 {width} {height}" '
        f'role="img" aria-label="Bar chart">'
        + axis
        + "".join(bars)
        + "".join(labels)
        + "</svg>"
    )


def render_stats_html(stats_data: Dict, title: str = "Phrase Stats Report") -> str:
    summary = stats_data.get("summary", {})
    frequency = stats_data.get("frequency", {})
    length = stats_data.get("length", {})
    diversity = stats_data.get("diversity", {})
    top_phrases = stats_data.get("top_phrases", [])

    bar_chart = _svg_bar_chart(top_phrases, "phrase", "frequency")

    return f"""<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <title>{escape(title)}</title>
  <style>
    :root {{
      --bg: #f6f2ea;
      --panel: #fffaf2;
      --ink: #2f2a24;
      --accent: #2a9d8f;
      --muted: #7a6f63;
    }}
    body {{
      margin: 0;
      font-family: "Spectral", "Georgia", serif;
      background: radial-gradient(circle at top, #f2e6d8 0%, var(--bg) 60%);
      color: var(--ink);
    }}
    main {{
      max-width: 980px;
      margin: 40px auto;
      padding: 24px;
    }}
    h1 {{
      font-size: 32px;
      margin: 0 0 16px 0;
      letter-spacing: 0.5px;
    }}
    .panel {{
      background: var(--panel);
      border: 1px solid #e7d7c6;
      border-radius: 14px;
      padding: 18px;
      margin-bottom: 20px;
      box-shadow: 0 10px 24px rgba(60, 45, 30, 0.12);
    }}
    .stats-grid {{
      display: grid;
      grid-template-columns: repeat(auto-fit, minmax(180px, 1fr));
      gap: 12px;
    }}
    .stat {{
      background: #fff;
      border-radius: 12px;
      padding: 10px 12px;
      border: 1px solid #efe1d2;
    }}
    .stat h3 {{
      margin: 0 0 6px 0;
      font-size: 12px;
      color: var(--muted);
      text-transform: uppercase;
      letter-spacing: 0.08em;
    }}
    .stat p {{
      margin: 0;
      font-size: 18px;
      font-weight: 600;
    }}
    .chart {{
      overflow-x: auto;
    }}
  </style>
</head>
<body>
  <main>
    <h1>{escape(title)}</h1>
    <div class="panel">
      <div class="stats-grid">
        <div class="stat"><h3>Total Phrases</h3><p>{summary.get("total_phrases", 0)}</p></div>
        <div class="stat"><h3>Text Lines</h3><p>{summary.get("text_lines", 0)}</p></div>
        <div class="stat"><h3>Frequency Mean</h3><p>{frequency.get("mean", 0):.2f}</p></div>
        <div class="stat"><h3>Length Mean</h3><p>{length.get("mean", 0):.2f}</p></div>
        <div class="stat"><h3>Entropy</h3><p>{diversity.get("entropy", 0):.3f}</p></div>
      </div>
    </div>
    <div class="panel">
      <h2>Top Phrases</h2>
      <div class="chart">{bar_chart}</div>
    </div>
  </main>
</body>
</html>
"""



"""
テキスト間の類似度分析モジュール

複数ファイル/テキスト間の類似度を計算し、コピペ検出や重複分析を行います。
"""

__author__ = "Takeshi SHIMIZU"
__copyright__ = "Copyright 2023"

import numpy as np
import pandas as pd
from typing import List, Dict, Tuple, Optional, Union
import logging
from pathlib import Path

logger = logging.getLogger(__name__)


class SimilarityAnalyzer:
    """
    複数テキスト間の類似度を分析するクラス

    複数の類似度計算手法を提供：
    - 'levenshtein': レーベンシュタイン距離（正確だが遅い）
    - 'jaccard': N-gram Jaccard係数（高速）
    - 'cosine': TF-IDFコサイン類似度（長文向け）
    - 'auto': 自動選択（デフォルト、実用性重視）

    使用例:
        >>> analyzer = SimilarityAnalyzer(method='auto')
        >>> matrix = analyzer.compare_files(['doc1.txt', 'doc2.txt', 'doc3.txt'])
        >>> pairs = analyzer.find_similar_pairs(matrix, threshold=0.7)
    """

    def __init__(self, method: str = 'auto', ngram_size: int = 3):
        """
        Parameters:
            method (str): 類似度計算手法
                - 'auto': 自動選択（デフォルト）
                - 'levenshtein': レーベンシュタイン距離
                - 'jaccard': N-gram Jaccard係数
                - 'cosine': TF-IDFコサイン類似度
            ngram_size (int): N-gramのサイズ（jaccard使用時）
        """
        self.method = method
        self.ngram_size = ngram_size
        self._validate_method()

    def _validate_method(self):
        """メソッドの妥当性を検証"""
        valid_methods = ['auto', 'levenshtein', 'jaccard', 'cosine']
        if self.method not in valid_methods:
            raise ValueError(
                f"method は {valid_methods} のいずれかである必要があります。"
                f"指定された値: {self.method}"
            )

    def _select_method(self, text_length: int) -> str:
        """
        テキスト長に応じて最適な手法を自動選択

        Parameters:
            text_length (int): テキストの平均文字数

        Returns:
            str: 選択された手法名
        """
        if self.method != 'auto':
            return self.method

        # 実用性重視の自動選択
        if text_length < 500:
            # 短文: レーベンシュタイン距離（正確）
            return 'levenshtein'
        elif text_length < 5000:
            # 中文: Jaccard係数（速度と精度のバランス）
            return 'jaccard'
        else:
            # 長文: コサイン類似度（高速）
            return 'cosine'

    def similarity_levenshtein(self, text1: str, text2: str) -> float:
        """
        レーベンシュタイン距離ベースの類似度

        Parameters:
            text1 (str): テキスト1
            text2 (str): テキスト2

        Returns:
            float: 類似度（0.0-1.0）
        """
        import Levenshtein
        distance = Levenshtein.distance(text1, text2)

        max_len = max(len(text1), len(text2))
        if max_len == 0:
            return 1.0

        similarity = 1.0 - (distance / max_len)
        return similarity

    def _levenshtein_pure_python(self, s1: str, s2: str) -> int:
        """
        純粋Pythonでのレーベンシュタイン距離計算（フォールバック用）

        Parameters:
            s1 (str): 文字列1
            s2 (str): 文字列2

        Returns:
            int: レーベンシュタイン距離
        """
        if len(s1) < len(s2):
            return self._levenshtein_pure_python(s2, s1)

        if len(s2) == 0:
            return len(s1)

        previous_row = range(len(s2) + 1)
        for i, c1 in enumerate(s1):
            current_row = [i + 1]
            for j, c2 in enumerate(s2):
                insertions = previous_row[j + 1] + 1
                deletions = current_row[j] + 1
                substitutions = previous_row[j] + (c1 != c2)
                current_row.append(min(insertions, deletions, substitutions))
            previous_row = current_row

        return previous_row[-1]

    def similarity_jaccard(self, text1: str, text2: str) -> float:
        """
        N-gram Jaccard係数ベースの類似度

        Parameters:
            text1 (str): テキスト1
            text2 (str): テキスト2

        Returns:
            float: 類似度（0.0-1.0）
        """
        ngrams1 = self._get_ngrams(text1, self.ngram_size)
        ngrams2 = self._get_ngrams(text2, self.ngram_size)

        if not ngrams1 and not ngrams2:
            return 1.0
        if not ngrams1 or not ngrams2:
            return 0.0

        intersection = len(ngrams1 & ngrams2)
        union = len(ngrams1 | ngrams2)

        return intersection / union if union > 0 else 0.0

    def _get_ngrams(self, text: str, n: int) -> set:
        """
        テキストからN-gramを抽出

        Parameters:
            text (str): テキスト
            n (int): N-gramのサイズ

        Returns:
            set: N-gramの集合
        """
        if len(text) < n:
            return {text} if text else set()

        return {text[i:i+n] for i in range(len(text) - n + 1)}

    def similarity_cosine(self, text1: str, text2: str) -> float:
        """
        TF-IDFコサイン類似度

        Parameters:
            text1 (str): テキスト1
            text2 (str): テキスト2

        Returns:
            float: 類似度（0.0-1.0）
        """
        try:
            from sklearn.feature_extraction.text import TfidfVectorizer
            from sklearn.metrics.pairwise import cosine_similarity

            vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(2, 3))
            vectors = vectorizer.fit_transform([text1, text2])
            similarity = cosine_similarity(vectors[0:1], vectors[1:2])[0][0]

            return float(similarity)

        except ImportError:
            logger.warning(
                "scikit-learn が見つかりません。jaccard法にフォールバックします。"
                "TF-IDF使用には 'pip install scikit-learn' が必要です。"
            )
            return self.similarity_jaccard(text1, text2)

    def calculate_similarity(self, text1: str, text2: str, method: Optional[str] = None) -> float:
        """
        2つのテキスト間の類似度を計算

        Parameters:
            text1 (str): テキスト1
            text2 (str): テキスト2
            method (str, optional): 使用する手法（Noneの場合はインスタンス設定を使用）

        Returns:
            float: 類似度（0.0-1.0）
        """
        if method is None:
            avg_length = (len(text1) + len(text2)) / 2
            method = self._select_method(int(avg_length))

        if method == 'levenshtein':
            return self.similarity_levenshtein(text1, text2)
        elif method == 'jaccard':
            return self.similarity_jaccard(text1, text2)
        elif method == 'cosine':
            return self.similarity_cosine(text1, text2)
        else:
            raise ValueError(f"不明な手法: {method}")

    def compare_texts(
        self,
        texts: List[str],
        labels: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """
        複数テキスト間の類似度行列を計算

        Parameters:
            texts (List[str]): テキストのリスト
            labels (List[str], optional): 各テキストのラベル

        Returns:
            pd.DataFrame: 類似度行列
        """
        n = len(texts)

        if labels is None:
            labels = [f"text_{i+1}" for i in range(n)]

        if len(labels) != n:
            raise ValueError(
                f"labels の数({len(labels)})が texts の数({n})と一致しません"
            )

        # 平均テキスト長を計算して手法を決定
        avg_length = sum(len(t) for t in texts) / n
        selected_method = self._select_method(int(avg_length))

        logger.info(f"類似度計算手法: {selected_method} (平均文字数: {int(avg_length)})")

        # 類似度行列を計算
        matrix = np.zeros((n, n))

        for i in range(n):
            for j in range(i, n):
                if i == j:
                    similarity = 1.0
                else:
                    similarity = self.calculate_similarity(
                        texts[i], texts[j], method=selected_method
                    )

                matrix[i, j] = similarity
                matrix[j, i] = similarity

        # DataFrameに変換
        df_matrix = pd.DataFrame(matrix, index=labels, columns=labels)

        return df_matrix

    def compare_files(
        self,
        filepaths: List[str],
        encoding: str = 'auto'
    ) -> pd.DataFrame:
        """
        複数ファイル間の類似度行列を計算

        Parameters:
            filepaths (List[str]): ファイルパスのリスト
            encoding (str): 文字エンコーディング

        Returns:
            pd.DataFrame: 類似度行列
        """
        from .utils import read_file

        texts = []
        labels = []

        for filepath in filepaths:
            text_lines = read_file(filepath, encoding=encoding)
            text = '\n'.join(text_lines)
            texts.append(text)

            # ファイル名をラベルとして使用
            label = Path(filepath).name
            labels.append(label)

        return self.compare_texts(texts, labels)

    def find_similar_pairs(
        self,
        similarity_matrix: pd.DataFrame,
        threshold: float = 0.7,
        top_n: Optional[int] = None
    ) -> List[Dict[str, Union[str, float]]]:
        """
        類似度の高いペアを抽出

        Parameters:
            similarity_matrix (pd.DataFrame): 類似度行列
            threshold (float): 類似度の閾値（0.0-1.0）
            top_n (int, optional): 上位N件のみ返す

        Returns:
            List[Dict]: 類似ペアのリスト
                [{'item1': 'text_1', 'item2': 'text_2', 'similarity': 0.85}, ...]
        """
        pairs = []
        labels = similarity_matrix.index.tolist()
        n = len(labels)

        for i in range(n):
            for j in range(i + 1, n):
                similarity = similarity_matrix.iloc[i, j]

                if similarity >= threshold:
                    pairs.append({
                        'item1': labels[i],
                        'item2': labels[j],
                        'similarity': float(similarity)
                    })

        # 類似度の降順でソート
        pairs.sort(key=lambda x: x['similarity'], reverse=True)

        if top_n is not None:
            pairs = pairs[:top_n]

        return pairs

    def export_matrix(
        self,
        similarity_matrix: pd.DataFrame,
        filepath: str,
        format: str = 'csv'
    ):
        """
        類似度行列をファイルに出力

        Parameters:
            similarity_matrix (pd.DataFrame): 類似度行列
            filepath (str): 出力先ファイルパス
            format (str): 出力形式（'csv', 'excel', 'json'）
        """
        if format == 'csv':
            similarity_matrix.to_csv(filepath, encoding='utf-8-sig')
        elif format == 'excel':
            similarity_matrix.to_excel(filepath, engine='openpyxl')
        elif format == 'json':
            similarity_matrix.to_json(filepath, orient='index', indent=2, force_ascii=False)
        else:
            raise ValueError(f"未対応の形式: {format}")

        logger.info(f"類似度行列を保存しました: {filepath}")

    def export_heatmap(
        self,
        similarity_matrix: pd.DataFrame,
        filepath: str,
        figsize: Tuple[int, int] = (10, 8),
        cmap: str = 'RdYlGn'
    ):
        """
        類似度行列をヒートマップとして出力

        Parameters:
            similarity_matrix (pd.DataFrame): 類似度行列
            filepath (str): 出力先ファイルパス
            figsize (Tuple[int, int]): 図のサイズ
            cmap (str): カラーマップ
        """
        try:
            import matplotlib.pyplot as plt
            import seaborn as sns

            plt.figure(figsize=figsize)
            sns.heatmap(
                similarity_matrix,
                annot=True,
                fmt='.2f',
                cmap=cmap,
                vmin=0,
                vmax=1,
                square=True,
                cbar_kws={'label': '類似度'}
            )
            plt.title('テキスト類似度行列')
            plt.tight_layout()
            plt.savefig(filepath, dpi=300, bbox_inches='tight')
            plt.close()

            logger.info(f"ヒートマップを保存しました: {filepath}")

        except ImportError as e:
            logger.error(
                f"ヒートマップ出力にはmatplotlibとseabornが必要です: {e}\n"
                "インストール: pip install matplotlib seaborn"
            )
            raise



# coding: utf-8
"""
分布比較エンジン（Distribution Comparator）

テキスト間・Part間・キャラ間の語彙分布の「違い」を精密測定する。
χ² 一本足からの脱却。

指標:
- Log-Likelihood Ratio (G²): スパースに強いキーネス分析
- Jensen-Shannon Divergence: 対称的な分布距離
- Effect Size (Log Ratio): 有意性ではなく「どれだけ違うか」
- Dice Coefficient: コーパスサイズ非依存の結合度
- Keyness Profiler: G² + Effect Size の統合キーワード抽出
"""

__author__ = "Takeshi SHIMIZU"
__copyright__ = "Copyright 2023-2026"

import math
import logging
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple
from collections import Counter

import numpy as np
import pandas as pd
from scipy import stats as sp_stats

logger = logging.getLogger(__name__)


@dataclass
class KeynessResult:
    """キーネス分析結果（1語分）"""
    term: str
    freq_target: int
    freq_reference: int
    total_target: int
    total_reference: int

    # 統計指標
    log_likelihood: float       # G² 値
    p_value: float              # G² の p 値
    log_ratio: float            # Effect Size (log2)
    dice_coefficient: float     # Dice 係数
    jsd_contribution: float     # JSD への寄与度

    # 判定
    significance_level: str     # ***, **, *, ns
    direction: str              # 'overuse' / 'underuse'
    keyness_score: float        # 複合キーネススコア (0-1)


@dataclass
class DistributionComparison:
    """2テキスト間の分布比較結果"""
    jsd: float                              # Jensen-Shannon Divergence
    jsd_sqrt: float                         # √JSD = Jensen-Shannon Distance
    total_target: int
    total_reference: int
    unique_target: int
    unique_reference: int
    shared_terms: int
    keyness_results: List[KeynessResult] = field(default_factory=list)


class DistributionComparator:
    """
    テキスト間の語彙分布比較エンジン

    使用例:
        >>> comp = DistributionComparator()
        >>> freq_a = Counter(["彼女", "彼女", "騎士", "剣", "剣", "剣"])
        >>> freq_b = Counter(["研究", "研究", "実験", "実験", "実験", "データ"])
        >>> result = comp.compare(freq_a, freq_b)
        >>> print(f"JSD: {result.jsd:.4f}")
    """

    def __init__(self, alpha: float = 0.05, min_freq: int = 2):
        """
        Parameters:
            alpha: 有意水準
            min_freq: キーネス分析の最小頻度（target + reference の合計）
        """
        self.alpha = alpha
        self.min_freq = min_freq

    # ─── Core Metrics ───────────────────────────────────────

    @staticmethod
    def log_likelihood_ratio(
        o11: int, o12: int, o21: int, o22: int
    ) -> float:
        """
        Log-Likelihood Ratio (G²)

        2×2 分割表:
            |  target  | reference |
        term|   o11    |   o12     |
        rest|   o21    |   o22     |

        G² = 2 * Σ O * ln(O/E)
        χ² よりスパースデータに強い。キーネス分析の標準指標。
        """
        table = np.array([[o11, o12], [o21, o22]], dtype=np.float64)
        # ゼロ対策
        table = np.maximum(table, 1e-10)
        total = table.sum()
        row_sums = table.sum(axis=1, keepdims=True)
        col_sums = table.sum(axis=0, keepdims=True)
        expected = row_sums * col_sums / total
        expected = np.maximum(expected, 1e-10)

        g2 = 2.0 * np.sum(table * np.log(table / expected))
        return float(g2)

    @staticmethod
    def log_ratio(freq_target: int, freq_reference: int,
                  total_target: int, total_reference: int) -> float:
        """
        Effect Size: Log Ratio (log2)

        log2(normalized_freq_target / normalized_freq_reference)
        正: target で過剰使用 / 負: target で過少使用
        """
        # ゼロ補正（Laplace smoothing）
        norm_t = (freq_target + 0.5) / (total_target + 1)
        norm_r = (freq_reference + 0.5) / (total_reference + 1)
        return math.log2(norm_t / norm_r)

    @staticmethod
    def dice_coefficient(freq_target: int, freq_reference: int,
                         total_target: int, total_reference: int) -> float:
        """
        Dice Coefficient

        2 * min(norm_t, norm_r) / (norm_t + norm_r)
        コーパスサイズに依存しない重複度。0-1。
        """
        norm_t = freq_target / max(total_target, 1)
        norm_r = freq_reference / max(total_reference, 1)
        denom = norm_t + norm_r
        if denom == 0:
            return 0.0
        return 2.0 * min(norm_t, norm_r) / denom

    @staticmethod
    def jensen_shannon_divergence(
        p: np.ndarray, q: np.ndarray
    ) -> float:
        """
        Jensen-Shannon Divergence

        JSD(P||Q) = 0.5 * KL(P||M) + 0.5 * KL(Q||M)
        where M = 0.5 * (P + Q)

        対称的・有界（0-1 when using log2）。分布間距離の標準。
        """
        # ゼロ補正
        p = np.asarray(p, dtype=np.float64)
        q = np.asarray(q, dtype=np.float64)

        # 正規化
        p_sum = p.sum()
        q_sum = q.sum()
        if p_sum == 0 or q_sum == 0:
            return 1.0  # 完全に異なる

        p = p / p_sum
        q = q / q_sum

        m = 0.5 * (p + q)

        # KL divergence（ゼロ回避）
        def _kl(a: np.ndarray, b: np.ndarray) -> float:
            mask = a > 0
            return float(np.sum(a[mask] * np.log2(a[mask] / b[mask])))

        jsd = 0.5 * _kl(p, m) + 0.5 * _kl(q, m)
        return max(0.0, min(1.0, jsd))

    # ─── High-Level API ─────────────────────────────────────

    def compare(
        self,
        freq_target: Counter,
        freq_reference: Counter,
        label_target: str = "target",
        label_reference: str = "reference",
    ) -> DistributionComparison:
        """
        2つの頻度分布を包括的に比較

        Parameters:
            freq_target: ターゲットテキストの語彙頻度
            freq_reference: リファレンステキストの語彙頻度
            label_target: ターゲットのラベル
            label_reference: リファレンスのラベル

        Returns:
            DistributionComparison
        """
        total_t = sum(freq_target.values())
        total_r = sum(freq_reference.values())

        # 全語彙の和集合
        all_terms = set(freq_target.keys()) | set(freq_reference.keys())
        shared = set(freq_target.keys()) & set(freq_reference.keys())

        # JSD 計算用ベクトル
        terms_list = sorted(all_terms)
        p_vec = np.array([freq_target.get(t, 0) for t in terms_list], dtype=np.float64)
        q_vec = np.array([freq_reference.get(t, 0) for t in terms_list], dtype=np.float64)

        jsd = self.jensen_shannon_divergence(p_vec, q_vec)

        # JSD 寄与度（各語の寄与を計算）
        p_norm = p_vec / max(p_vec.sum(), 1)
        q_norm = q_vec / max(q_vec.sum(), 1)
        m_norm = 0.5 * (p_norm + q_norm)

        jsd_contributions = {}
        for i, term in enumerate(terms_list):
            contrib = 0.0
            if p_norm[i] > 0 and m_norm[i] > 0:
                contrib += 0.5 * p_norm[i] * math.log2(p_norm[i] / m_norm[i])
            if q_norm[i] > 0 and m_norm[i] > 0:
                contrib += 0.5 * q_norm[i] * math.log2(q_norm[i] / m_norm[i])
            jsd_contributions[term] = max(0.0, contrib)

        # キーネス分析（全語に対して）
        keyness_results = []
        for term in terms_list:
            ft = freq_target.get(term, 0)
            fr = freq_reference.get(term, 0)

            # 最小頻度フィルタ
            if ft + fr < self.min_freq:
                continue

            # 2×2 分割表
            o11 = ft
            o12 = fr
            o21 = total_t - ft
            o22 = total_r - fr

            g2 = self.log_likelihood_ratio(o11, o12, o21, o22)
            lr = self.log_ratio(ft, fr, total_t, total_r)
            dice = self.dice_coefficient(ft, fr, total_t, total_r)

            # p 値（G² は自由度1のχ²分布に近似）
            p_value = 1.0 - sp_stats.chi2.cdf(abs(g2), df=1)

            # 有意水準判定
            if p_value < 0.001:
                sig = "***"
            elif p_value < 0.01:
                sig = "**"
            elif p_value < self.alpha:
                sig = "*"
            else:
                sig = "ns"

            direction = "overuse" if lr > 0 else "underuse"

            # 複合キーネススコア
            # G² の正規化 + |log_ratio| + JSD寄与
            g2_norm = min(1.0, g2 / 100.0)  # G²=100 で飽和
            lr_norm = min(1.0, abs(lr) / 5.0)  # |LR|=5 で飽和
            jsd_c = jsd_contributions.get(term, 0.0)
            jsd_c_norm = min(1.0, jsd_c / 0.01)  # 0.01 で飽和

            keyness_score = (
                0.45 * g2_norm +
                0.35 * lr_norm +
                0.20 * jsd_c_norm
            )

            keyness_results.append(KeynessResult(
                term=term,
                freq_target=ft,
                freq_reference=fr,
                total_target=total_t,
                total_reference=total_r,
                log_likelihood=g2,
                p_value=p_value,
                log_ratio=lr,
                dice_coefficient=dice,
                jsd_contribution=jsd_c,
                significance_level=sig,
                direction=direction,
                keyness_score=keyness_score,
            ))

        # キーネススコア降順
        keyness_results.sort(key=lambda r: r.keyness_score, reverse=True)

        return DistributionComparison(
            jsd=jsd,
            jsd_sqrt=math.sqrt(jsd),
            total_target=total_t,
            total_reference=total_r,
            unique_target=len(freq_target),
            unique_reference=len(freq_reference),
            shared_terms=len(shared),
            keyness_results=keyness_results,
        )

    def compare_to_dataframe(
        self,
        freq_target: Counter,
        freq_reference: Counter,
        top_k: Optional[int] = None,
    ) -> pd.DataFrame:
        """比較結果を DataFrame で返す"""
        result = self.compare(freq_target, freq_reference)
        rows = []
        for kr in result.keyness_results:
            rows.append({
                'term': kr.term,
                'freq_target': kr.freq_target,
                'freq_reference': kr.freq_reference,
                'log_likelihood': kr.log_likelihood,
                'p_value': kr.p_value,
                'log_ratio': kr.log_ratio,
                'dice_coefficient': kr.dice_coefficient,
                'jsd_contribution': kr.jsd_contribution,
                'significance': kr.significance_level,
                'direction': kr.direction,
                'keyness_score': kr.keyness_score,
            })

        df = pd.DataFrame(rows)
        if top_k and not df.empty:
            df = df.head(top_k)
        return df

    def generate_report(
        self,
        freq_target: Counter,
        freq_reference: Counter,
        label_target: str = "Target",
        label_reference: str = "Reference",
        top_k: int = 20,
    ) -> str:
        """比較レポート生成"""
        result = self.compare(freq_target, freq_reference, label_target, label_reference)

        lines = [
            "=" * 70,
            f"【分布比較レポート】 {label_target} vs {label_reference}",
            "=" * 70,
            "",
            f"Jensen-Shannon Divergence: {result.jsd:.6f}",
            f"Jensen-Shannon Distance:   {result.jsd_sqrt:.6f}",
            f"  ({label_target}: {result.total_target}語 / {label_reference}: {result.total_reference}語)",
            f"  ユニーク語彙: {result.unique_target} / {result.unique_reference}",
            f"  共有語彙: {result.shared_terms}",
            "",
        ]

        # JSD の解釈
        if result.jsd < 0.05:
            lines.append("  → 分布は非常に類似しています")
        elif result.jsd < 0.15:
            lines.append("  → 分布にやや差があります")
        elif result.jsd < 0.30:
            lines.append("  → 分布に明確な差があります")
        else:
            lines.append("  → 分布は大きく異なります")

        lines.append("")
        lines.append("-" * 70)

        # Overuse TOP
        overuse = [kr for kr in result.keyness_results if kr.direction == "overuse"]
        underuse = [kr for kr in result.keyness_results if kr.direction == "underuse"]

        lines.append(f"【{label_target} で過剰使用 TOP {min(top_k, len(overuse))}】")
        for i, kr in enumerate(overuse[:top_k], 1):
            lines.append(
                f"  {i:2d}. {kr.term:<12s} "
                f"G²={kr.log_likelihood:7.2f} {kr.significance_level:>3s}  "
                f"LR={kr.log_ratio:+.2f}  "
                f"({kr.freq_target}/{kr.freq_reference})"
            )

        lines.append("")
        lines.append(f"【{label_target} で過少使用 TOP {min(top_k, len(underuse))}】")
        for i, kr in enumerate(underuse[:top_k], 1):
            lines.append(
                f"  {i:2d}. {kr.term:<12s} "
                f"G²={kr.log_likelihood:7.2f} {kr.significance_level:>3s}  "
                f"LR={kr.log_ratio:+.2f}  "
                f"({kr.freq_target}/{kr.freq_reference})"
            )

        lines.append("")
        lines.append("=" * 70)
        return "\n".join(lines)

    # ─── Utilities ──────────────────────────────────────────

    @staticmethod
    def text_to_freq(text: str, min_length: int = 2, max_length: int = 10) -> Counter:
        """
        テキストから N-gram 頻度カウンターを生成（簡易版）

        PhraseExtracter と併用する場合は不要。
        単独で使いたいときの便利メソッド。
        """
        import re
        freq = Counter()
        # 空白・改行で分割しない文字 N-gram
        clean = re.sub(r'\s+', '', text)
        for n in range(min_length, max_length + 1):
            for i in range(len(clean) - n + 1):
                gram = clean[i:i + n]
                freq[gram] += 1
        return freq

    @staticmethod
    def dataframe_to_freq(
        df: pd.DataFrame,
        phrase_col: str = 'phrase',
        freq_col: str = 'freq',
    ) -> Counter:
        """DataFrame を Counter に変換"""
        # 列名の自動検出
        if phrase_col not in df.columns:
            phrase_col = 'seqchar' if 'seqchar' in df.columns else df.columns[0]
        if freq_col not in df.columns:
            freq_col = 'frequency' if 'frequency' in df.columns else 'count'

        return Counter(dict(zip(df[phrase_col], df[freq_col])))



"""
高度なエンコーディング検出と修復

複雑な文字化けへの対応、BOMの処理、複数エンコーディング候補の試行
文字化けに対する堅牢な対応を提供
"""

import sys
from typing import Tuple, List, Dict, Optional
import logging
import re

try:
    import chardet
    HAS_CHARDET = True
except ImportError:
    HAS_CHARDET = False

logger = logging.getLogger(__name__)


class EncodingDetector:
    """高度なエンコーディング検出と修復"""

    # よく使われるエンコーディング候補（優先順位順）
    ENCODING_CANDIDATES = [
        'utf-8',
        'utf-8-sig',  # BOM付きUTF-8
        'cp932',  # Shift-JIS（Windows標準）
        'shift_jis',
        'euc-jp',
        'iso-2022-jp',
        'ascii',
    ]

    # BOMシグネチャ
    BOM_SIGNATURES = {
        b'\xff\xfe\x00\x00': 'utf-32-le',
        b'\x00\x00\xfe\xff': 'utf-32-be',
        b'\xff\xfe': 'utf-16-le',
        b'\xfe\xff': 'utf-16-be',
        b'\xef\xbb\xbf': 'utf-8-sig',
    }

    # 文字化けパターン（修復パターン）
    GARBLED_PATTERNS = {
        # UTF-8をShift-JISで読んだ場合
        'utf8_as_sjis': {
            'pattern': r'[\u0080-\u009f\u00a0-\u00ff]{2,}',  # 制御文字が多い
            'source_enc': 'cp932',
            'target_enc': 'utf-8'
        },
        # Shift-JISをUTF-8で読んだ場合
        'sjis_as_utf8': {
            'pattern': r'[\ufffd]+',  # Replacement character
            'source_enc': 'utf-8',
            'target_enc': 'cp932'
        },
    }

    def __init__(self):
        self.last_detection = None
        self.last_confidence = 0.0
        self.trials = []  # 試行ログ

    @staticmethod
    def detect_bom(data: bytes) -> Optional[str]:
        """BOM（Byte Order Mark）を検出

        Returns:
            BOMが見つかった場合はエンコーディング名、なければNone
        """
        for bom, encoding in EncodingDetector.BOM_SIGNATURES.items():
            if data.startswith(bom):
                logger.debug(f"BOM detected: {encoding}")
                return encoding
        return None

    @staticmethod
    def detect_with_fallback(data: bytes, try_repair: bool = True) -> Tuple[str, float]:
        """複数の候補からエンコーディングを検出（文字化け対応）

        Args:
            data: バイト列
            try_repair: 文字化けの修復を試みるかどうか

        Returns:
            (エンコーディング名, 信頼度)
        """
        # BOMを確認
        bom_encoding = EncodingDetector.detect_bom(data)
        if bom_encoding:
            return bom_encoding, 1.0

        # chardet が利用可能なら使用
        if HAS_CHARDET:
            result = chardet.detect(data)
            if result and result.get('encoding') and result.get('confidence', 0) >= 0.7:
                encoding = result['encoding']
                confidence = result['confidence']

                # Shift-JIS系を正規化
                if encoding and ('shift' in encoding.lower() or 'cp932' in encoding.lower() or 'ms932' in encoding.lower()):
                    encoding = 'cp932'

                logger.debug(f"chardet detected: {encoding} (confidence: {confidence})")
                return encoding, confidence

        # 候補を順番に試す（文字化けのシグナルをチェック）
        best_encoding = None
        best_score = -1

        for encoding in EncodingDetector.ENCODING_CANDIDATES:
            try:
                text = data.decode(encoding)

                # デコード成功時、文字化けスコアを計算
                garble_score = EncodingDetector._calculate_garble_score(text)
                logger.debug(f"{encoding}: garble_score={garble_score:.2f}")

                if garble_score > best_score:
                    best_score = garble_score
                    best_encoding = encoding

            except (UnicodeDecodeError, LookupError) as e:
                logger.debug(f"{encoding}: decode failed - {e}")
                continue

        if best_encoding:
            confidence = max(0.5, 1.0 - best_score / 10.0)  # スコアを信頼度に変換
            logger.debug(f"Best match: {best_encoding} (score: {best_score:.2f})")
            return best_encoding, confidence

        # 最終手段：UTF-8 with errors='replace'
        logger.warning("All encoding attempts failed, using UTF-8 with error handling")
        return 'utf-8', 0.1

    @staticmethod
    def _calculate_garble_score(text: str) -> float:
        """テキストの文字化けスコアを計算

        Returns:
            スコア（低いほど文字化けの可能性が低い）
        """
        if not text:
            return 0.0

        score = 0.0

        # 置換文字（U+FFFD）の出現
        replacement_count = text.count('\ufffd')
        score += replacement_count * 5.0

        # 制御文字の出現
        control_chars = sum(1 for c in text if ord(c) < 0x20 and c not in '\n\r\t')
        score += control_chars * 2.0

        # 不正なUnicode範囲
        invalid_chars = sum(1 for c in text if ord(c) in range(0xD800, 0xE000))
        score += invalid_chars * 3.0

        # 疑わしい文字パターン（例：化けたカナ）
        suspicious_patterns = len(re.findall(r'[\u0080-\u009f]{2,}', text))
        score += suspicious_patterns * 1.5

        return score

    @staticmethod
    def try_repair_garbled_text(data: bytes, assumed_encoding: str = 'utf-8') -> Tuple[str, str, bool]:
        """文字化けを修復

        Args:
            data: バイト列
            assumed_encoding: 仮定されるエンコーディング

        Returns:
            (テキスト, 使用エンコーディング, 修復したかどうか)
        """
        # 1. 仮定されるエンコーディングで試す
        try:
            text = data.decode(assumed_encoding)
            garble_score = EncodingDetector._calculate_garble_score(text)

            if garble_score < 1.0:  # 文字化けしていない
                return text, assumed_encoding, False
        except UnicodeDecodeError:
            pass

        # 2. 文字化け修復パターンを試す
        for pattern_name, pattern_info in EncodingDetector.GARBLED_PATTERNS.items():
            try:
                # 元のバイト列を別のエンコーディングで読み直す
                text_attempt = data.decode(
                    pattern_info['source_enc'],
                    errors='replace'
                )

                # パターンにマッチするかチェック
                if re.search(pattern_info['pattern'], text_attempt):
                    # 修復を試みる
                    repaired = EncodingDetector.repair_encoding(
                        text_attempt,
                        pattern_info['source_enc'],
                        pattern_info['target_enc']
                    )

                    garble_score = EncodingDetector._calculate_garble_score(repaired)
                    logger.info(f"Garble repair attempt '{pattern_name}': score={garble_score:.2f}")

                    if garble_score < 1.0:
                        logger.info(f"Successfully repaired using pattern: {pattern_name}")
                        return repaired, pattern_info['target_enc'], True

            except Exception as e:
                logger.debug(f"Repair pattern '{pattern_name}' failed: {e}")
                continue

        # 3. すべて失敗時は、最もスコアの低いエンコーディングを選ぶ
        best_encoding = assumed_encoding
        best_score = float('inf')
        best_text = None

        for encoding in EncodingDetector.ENCODING_CANDIDATES:
            try:
                text = data.decode(encoding, errors='replace')
                score = EncodingDetector._calculate_garble_score(text)

                if score < best_score:
                    best_score = score
                    best_encoding = encoding
                    best_text = text

            except Exception:
                continue

        if best_text is not None:
            logger.info(f"Fallback to {best_encoding} (score: {best_score:.2f})")
            return best_text, best_encoding, True

        # 最後の手段
        text = data.decode('utf-8', errors='replace')
        return text, 'utf-8', True

    @staticmethod
    def repair_encoding(
        text: str,
        original_encoding: str,
        target_encoding: str = 'utf-8'
    ) -> str:
        """誤ったエンコーディングで読まれたテキストを修復

        例: "あいう"が"ぁぃぅ"に化けた場合を修復

        Args:
            text: 化けているテキスト
            original_encoding: 誤ったエンコーディング
            target_encoding: 目標エンコーディング
        """
        if original_encoding == target_encoding:
            return text

        try:
            # 元のバイト列に戻す
            bytes_data = text.encode(original_encoding, errors='replace')

            # 正しいエンコーディングで再度デコード
            repaired = bytes_data.decode(target_encoding, errors='replace')
            logger.info(f"Repaired encoding: {original_encoding} -> {target_encoding}")
            return repaired

        except Exception as e:
            logger.warning(f"Encoding repair failed: {e}")
            return text

    @staticmethod
    def detect_encoding_errors(text: str) -> List[Dict]:
        """疑わしい文字セットを検出

        Returns:
            文字化けの可能性がある位置と修復提案のリスト
        """
        errors = []
        suspicious_patterns = [
            # UTF-8の一部をcp932で読んだ場合のパターン
            (r'[\ufffd]+', 'Replacement character detected (encoding mismatch)'),
            # 化けたカナのパターン（半角カナが混在）
            (r'[ァ-ヴー]{1,}[^\w\n ァ-ヴー，。！？]', 'Suspicious katakana sequence'),
        ]

        for i, char in enumerate(text):
            code = ord(char)

            # 制御文字（0x00-0x1F, 0x7F-0x9F）を検出
            if (0x00 <= code <= 0x1F) or (0x7F <= code <= 0x9F):
                if char not in '\n\r\t':
                    errors.append({
                        'position': i,
                        'char': repr(char),
                        'code': code,
                        'type': 'control_character'
                    })

            # 不正なUnicode範囲
            elif code == 0xFFFD:  # Replacement character
                errors.append({
                    'position': i,
                    'char': char,
                    'type': 'replacement_character'
                })

        return errors

    @staticmethod
    def remove_control_characters(text: str) -> str:
        """制御文字を除去

        Args:
            text: テキスト

        Returns:
            制御文字が除去されたテキスト
        """
        return ''.join(
            char for char in text
            if (ord(char) >= 0x20 and ord(char) != 0x7F) or char in '\n\r\t'
        )


class BOMHandler:
    """BOM（Byte Order Mark）の処理"""

    @staticmethod
    def remove_bom(text: str, encoding: str) -> str:
        """テキストからBOMを除去

        Args:
            text: BOM付きテキスト
            encoding: エンコーディング

        Returns:
            BOMが除去されたテキスト
        """
        if encoding == 'utf-8-sig' and text.startswith('\ufeff'):
            return text[1:]
        elif text.startswith('\ufeff'):
            return text[1:]
        return text

    @staticmethod
    def add_bom(text: str, encoding: str) -> bytes:
        """テキストにBOMを付加してバイト列を返す

        Args:
            text: テキスト
            encoding: エンコーディング

        Returns:
            BOM付きバイト列
        """
        if encoding == 'utf-8-sig':
            return b'\xef\xbb\xbf' + text.encode('utf-8')
        elif encoding == 'utf-16':
            # UTF-16はデフォルトでBOM付き
            return text.encode('utf-16')
        else:
            return text.encode(encoding)

    @staticmethod
    def normalize_encoding(encoding: str) -> str:
        """エンコーディング名を正規化

        Args:
            encoding: エンコーディング名

        Returns:
            正規化されたエンコーディング名
        """
        encoding = encoding.lower().replace('-', '_').replace('_', '-')

        # Shift-JIS系を統一
        if 'shift' in encoding or 'sjis' in encoding or 'cp932' in encoding or 'ms932' in encoding:
            return 'cp932'

        # EUC-JP系を統一
        if 'euc' in encoding and 'jp' in encoding:
            return 'euc-jp'

        # ISO-2022-JP系を統一
        if 'iso' in encoding and '2022' in encoding and 'jp' in encoding:
            return 'iso-2022-jp'

        return encoding


class EncodingValidator:
    """エンコーディングの妥当性チェック"""

    @staticmethod
    def is_valid_utf8(data: bytes) -> bool:
        """データがUTF-8として有効かチェック"""
        try:
            data.decode('utf-8')
            return True
        except UnicodeDecodeError:
            return False

    @staticmethod
    def is_valid_encoding(data: bytes, encoding: str) -> bool:
        """データがspecified encodingとして有効かチェック"""
        try:
            data.decode(encoding)
            return True
        except (UnicodeDecodeError, LookupError):
            return False

    @staticmethod
    def validate_and_repair(data: bytes, encoding: str) -> Tuple[str, str]:
        """エンコーディングを検証し、失敗した場合は修復を試みる

        Args:
            data: バイト列
            encoding: 指定エンコーディング

        Returns:
            (デコードされたテキスト, 使用されたエンコーディング)
        """
        # 指定のエンコーディングで試す
        try:
            text = data.decode(encoding)
            logger.debug(f"Decoded successfully with {encoding}")
            return text, encoding
        except (UnicodeDecodeError, LookupError):
            logger.warning(f"Failed to decode with {encoding}, trying alternatives...")

        # 自動検出にフォールバック
        detected, _ = EncodingDetector.detect_with_fallback(data)
        try:
            text = data.decode(detected)
            logger.info(f"Successfully decoded with detected encoding: {detected}")
            return text, detected
        except UnicodeDecodeError:
            # 最終手段
            text = data.decode(detected, errors='replace')
            logger.warning(f"Using error handling with {detected}")
            return text, detected



# coding: utf-8
"""
ストリーミング・インクリメンタル処理

大規模テキスト（100GB+）をメモリ効率的に処理
"""

import logging
from typing import List, Dict, Tuple, Optional, Generator
from collections import defaultdict, Counter
import pandas as pd

logger = logging.getLogger(__name__)


class StreamingPhraseExtracter:
    """ストリーミング処理対応のフレーズ抽出"""

    def __init__(self, min_count: int = 3, min_length: int = 1, max_length: int = 10):
        """
        Parameters:
            min_count: 最小出現回数
            min_length: 最小フレーズ長
            max_length: 最大フレーズ長
        """
        self.min_count = min_count
        self.min_length = min_length
        self.max_length = max_length

        # ストリーミング状態
        self.phrase_counter = defaultdict(int)
        self.total_texts = 0
        self.chunks_processed = 0

    def add_chunk(self, texts: List[str]) -> Tuple[pd.DataFrame, Dict]:
        """
        新しいテキストチャンク（例：1000行）を追加

        Parameters:
            texts: テキストリスト

        Returns:
            (現在のフレーズ結果DataFrame, 処理統計)
        """
        from .extracter import PhraseExtracter

        # フレーズを抽出（ここから import されているはずなので使用）
        try:
            extractor = PhraseExtracter(
                min_count=1,  # 一時的に1でカウント
                min_length=self.min_length,
                max_length=self.max_length,
                verbose=0
            )
            df_chunk = extractor.extract(texts)
        except Exception as e:
            logger.warning(f"チャンク処理エラー: {e}")
            df_chunk = pd.DataFrame()

        if df_chunk is not None and not df_chunk.empty:
            phrase_col = 'seqchar' if 'seqchar' in df_chunk.columns else 'phrase'
            freq_col = 'freq' if 'freq' in df_chunk.columns else 'frequency'

            # フレーズカウンターを更新
            for idx, row in df_chunk.iterrows():
                phrase = row[phrase_col]
                freq = row[freq_col]
                self.phrase_counter[phrase] += freq

        # 処理統計
        self.total_texts += len(texts)
        self.chunks_processed += 1

        stats = {
            'total_texts': self.total_texts,
            'chunks_processed': self.chunks_processed,
            'unique_phrases': len(self.phrase_counter),
            'chunk_size': len(texts),
        }

        logger.info(
            f"チャンク {self.chunks_processed} 処理完了: "
            f"テキスト {len(texts)} 行, "
            f"ユニークフレーズ {len(self.phrase_counter)} 個"
        )

        return self.get_current_results(), stats

    def get_current_results(self, top_k: Optional[int] = None) -> pd.DataFrame:
        """
        現在のところの暫定結果を返す

        Parameters:
            top_k: 上位K件を返す（Noneなら全件）

        Returns:
            フレーズDataFrame
        """
        # min_count でフィルタ
        filtered = {
            phrase: count
            for phrase, count in self.phrase_counter.items()
            if count >= self.min_count
        }

        if not filtered:
            return pd.DataFrame(columns=['seqchar', 'freq'])

        # DataFrame に変換
        df = pd.DataFrame([
            {'seqchar': phrase, 'freq': count}
            for phrase, count in filtered.items()
        ])

        # 頻度でソート
        df = df.sort_values('freq', ascending=False)

        if top_k is not None:
            df = df.head(top_k)

        return df.reset_index(drop=True)

    def finalize(self) -> pd.DataFrame:
        """全チャンク処理後の最終結果を返す"""
        return self.get_current_results()

    def get_statistics(self) -> Dict:
        """ストリーミング処理の統計を取得"""
        total_count = sum(self.phrase_counter.values())

        return {
            'total_texts': self.total_texts,
            'chunks_processed': self.chunks_processed,
            'unique_phrases': len(self.phrase_counter),
            'total_phrase_count': total_count,
            'avg_chunk_size': self.total_texts / self.chunks_processed if self.chunks_processed > 0 else 0,
        }

    def memory_estimate(self) -> Dict:
        """メモリ使用量の推定"""
        # Counter のメモリ: フレーズ * (50 + フレーズ長)
        phrase_memory = sum(
            50 + len(phrase)
            for phrase in self.phrase_counter.keys()
        ) / (1024 * 1024)  # MB

        return {
            'phrase_counter_mb': phrase_memory,
            'estimated_total_mb': phrase_memory * 1.5,  # オーバーヘッド込み
        }


class ChunkedTextReader:
    """テキストファイルをチャンク単位で読み込み"""

    def __init__(self, filepath: str, chunk_size: int = 1000):
        """
        Parameters:
            filepath: テキストファイルパス
            chunk_size: チャンクサイズ（行数）
        """
        self.filepath = filepath
        self.chunk_size = chunk_size

    def read_chunks(self) -> Generator[List[str], None, None]:
        """
        チャンク単位でテキストを読み込み

        Yields:
            テキストリスト（chunk_size 行）
        """
        chunk = []

        try:
            with open(self.filepath, 'r', encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    if line:
                        chunk.append(line)

                    if len(chunk) >= self.chunk_size:
                        yield chunk
                        chunk = []

                # 残りのデータ
                if chunk:
                    yield chunk

        except Exception as e:
            logger.error(f"ファイル読み込みエラー: {e}")
            raise


class IncrementalAggregator:
    """インクリメンタル集約"""

    def __init__(self):
        self.aggregated = {}
        self.updates = []

    def add_phrase(self, phrase: str, count: int) -> None:
        """フレーズをカウンター"""
        if phrase not in self.aggregated:
            self.aggregated[phrase] = 0

        old_count = self.aggregated[phrase]
        new_count = old_count + count

        self.aggregated[phrase] = new_count

        # 更新履歴を記録（オプション）
        self.updates.append({
            'phrase': phrase,
            'old_count': old_count,
            'new_count': new_count,
            'delta': count,
        })

    def get_snapshot(self) -> Dict[str, int]:
        """現在のスナップショット"""
        return self.aggregated.copy()

    def get_changes_since(self, last_update: int = 0) -> Dict[str, int]:
        """最後の更新以降の変更"""
        recent_updates = self.updates[last_update:]

        changes = {}
        for update in recent_updates:
            phrase = update['phrase']
            if phrase not in changes:
                changes[phrase] = 0
            changes[phrase] += update['delta']

        return changes


class StreamingAnalyzer:
    """ストリーミング処理の統合分析"""

    def __init__(self, min_count: int = 3, chunk_size: int = 1000):
        self.extractor = StreamingPhraseExtracter(min_count=min_count)
        self.reader = None
        self.aggregator = IncrementalAggregator()
        self.chunk_size = chunk_size

    def process_file(
        self,
        filepath: str,
        callback: Optional[callable] = None
    ) -> Tuple[pd.DataFrame, Dict]:
        """
        ファイルをストリーミング処理

        Parameters:
            filepath: ファイルパス
            callback: 各チャンク処理後のコールバック関数
                    (df_results, stats) を受け取る

        Returns:
            (最終結果DataFrame, 処理統計)
        """
        self.reader = ChunkedTextReader(filepath, self.chunk_size)

        stats_total = {
            'total_chunks': 0,
            'total_texts': 0,
            'processing_started': True,
        }

        for chunk in self.reader.read_chunks():
            # チャンクを処理
            df_results, chunk_stats = self.extractor.add_chunk(chunk)

            # コールバック実行
            if callback:
                callback(df_results, chunk_stats)

            # 統計を集約
            stats_total['total_chunks'] += 1
            stats_total['total_texts'] += chunk_stats['total_texts']

            logger.info(
                f"進捗: チャンク {stats_total['total_chunks']}, "
                f"テキスト {stats_total['total_texts']} 行"
            )

        # 最終結果
        final_results = self.extractor.finalize()
        final_stats = self.extractor.get_statistics()

        return final_results, final_stats

    def generate_streaming_report(self) -> str:
        """ストリーミング処理のレポート"""
        stats = self.extractor.get_statistics()
        mem = self.extractor.memory_estimate()

        report = "【ストリーミング処理レポート】\n\n"
        report += f"処理済みテキスト: {stats['total_texts']} 行\n"
        report += f"処理チャンク数: {stats['chunks_processed']}\n"
        report += f"ユニークフレーズ: {stats['unique_phrases']} 個\n"
        report += f"総フレーズ数: {stats['total_phrase_count']}\n"
        report += f"平均チャンクサイズ: {stats['avg_chunk_size']:.1f} 行\n\n"

        report += "【メモリ使用量】\n"
        report += f"フレーズカウンター: {mem['phrase_counter_mb']:.2f} MB\n"
        report += f"推定総メモリ: {mem['estimated_total_mb']:.2f} MB\n"

        return report



"""
ドキュメント ベクトル化 ユーティリティ

NMF（Non-negative Matrix Factorization）ベースのドキュメント ベクトル化と
PMIフィルタリングに関するヘルパー関数を提供します。
"""

__author__ = "Takeshi SHIMIZU"
__copyright__ = "Copyright 2023"

import numpy as np
import pandas as pd
from typing import List, Tuple, Optional, Any
import logging
try:
    from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
    from sklearn.decomposition import NMF
    _HAS_SKLEARN = True
except ImportError:
    TfidfVectorizer = None
    CountVectorizer = None
    NMF = None
    _HAS_SKLEARN = False

logger = logging.getLogger(__name__)


def extract_pmi_filtered_phrases(
    texts: List[str],
    mode: str = 'low_pmi',
    pmi_threshold: float = 3.0,
    min_count: int = 6,
    min_length: int = 2,
    max_length: int = 15
) -> Optional[List[str]]:
    """
    PMIスコアでフィルタリングされたフレーズを抽出

    Parameters:
        texts (List[str]): テキストのリスト
        mode (str): フィルタリングモード
            - 'low_pmi': 低PMIフレーズのみ（習慣的表現）
            - 'high_pmi': 高PMIフレーズのみ（意味的表現）
            - 'all': フィルタリングなし
        pmi_threshold (float): PMI閾値
        min_count (int): 最小出現回数
        min_length (int): 最小フレーズ長
        max_length (int): 最大フレーズ長

    Returns:
        List[str]: フィルタリングされたフレーズリスト、またはNone（allモードの場合）
    """
    if mode == 'all':
        return None

    # テキストを結合
    combined_text = '\n'.join(texts)

    if mode == 'low_pmi':
        # WritingHabitDetectorを使用して低PMIフレーズを抽出
        from .writing_habit_detector import WritingHabitDetector

        detector = WritingHabitDetector(
            min_count=min_count,
            max_pmi=pmi_threshold,
            min_length=min_length,
            max_length=max_length,
            verbose=False
        )

        df = detector.detect(combined_text)
        if df.empty:
            logger.warning(f"Low PMI mode: フレーズが抽出されませんでした")
            return None

        if 'phrase' in df.columns:
            return df['phrase'].tolist()
        else:
            return None

    elif mode == 'high_pmi':
        # PhraseExtracterを使用して高PMIフレーズを抽出
        from .extracter import PhraseExtracter

        extractor = PhraseExtracter(
            min_count=min_count,
            min_length=min_length,
            max_length=max_length,
            use_pmi=True,
            verbose=0
        )

        sentences = combined_text.split('\n')
        df = extractor.get_dfphrase(sentences)

        if df.empty:
            logger.warning(f"High PMI mode: フレーズが抽出されませんでした")
            return None

        # PMIが高いフレーズのみ
        if 'pmi' in df.columns:
            df_filtered = df[df['pmi'] >= pmi_threshold]
            if df_filtered.empty:
                logger.warning(f"High PMI mode: PMI >= {pmi_threshold} のフレーズがありません")
                return None
            seqchar_col = 'seqchar' if 'seqchar' in df_filtered.columns else 'phrase'
            return df_filtered[seqchar_col].tolist()
        else:
            # PMIがない場合は全て返す
            seqchar_col = 'seqchar' if 'seqchar' in df.columns else 'phrase'
            return df[seqchar_col].tolist()

    else:
        raise ValueError(f"Unknown PMI mode: {mode}")


def build_document_term_matrix(
    texts: List[str],
    feature_mode: str = 'tfidf',
    pmi_threshold: float = 3.0,
    min_count: int = 6,
    max_features: int = 1000,
    ngram_range: Tuple[int, int] = (2, 3),
    min_df: int = 1,
    max_df: float = 0.95,
    analyzer: str = 'char',
    verbose: int = 1
) -> Tuple[np.ndarray, Any, List[str], dict]:
    """
    PMIベースのフィルタリングを適用したドキュメント-ターム行列を構築

    Pipeline:
    1. PMIフィルタリングで語彙を制限（必要な場合）
    2. TF-IDFまたはカウント行列を構築
    3. 行列と特徴名を返す

    Parameters:
        texts (List[str]): テキストのリスト
        feature_mode (str): 特徴抽出モード
            - 'tfidf': 通常のTF-IDF（意味的差）
            - 'low_pmi': 低PMIフレーズのみ（手癖・文体差）
            - 'high_pmi': 高PMIフレーズのみ（意味的表現差）
            - 'hybrid': TF-IDFと低PMIを組み合わせ
        pmi_threshold (float): PMI閾値
        min_count (int): 最小出現回数
        max_features (int): 最大特徴数
        ngram_range (Tuple[int, int]): N-gramの範囲
        min_df (int): ドキュメント最小出現数
        max_df (float): ドキュメント最大出現率
        analyzer (str): ベクトル化のアナライザー（char, word）
        verbose (int): ログレベル

    Returns:
        Tuple: (行列, ベクトライザー, 特徴名, メタデータ)
    """
    vocabulary = None
    metadata = {
        'feature_mode': feature_mode,
        'pmi_threshold': pmi_threshold,
        'min_count': min_count,
    }

    # ステップ1: PMIフィルタリングで語彙を制限（必要な場合）
    if feature_mode in ['low_pmi', 'high_pmi']:
        if verbose:
            logger.info(f"PMI filtering mode: {feature_mode}")

        vocabulary = extract_pmi_filtered_phrases(
            texts,
            mode=feature_mode,
            pmi_threshold=pmi_threshold,
            min_count=min_count
        )

        if vocabulary is None or len(vocabulary) == 0:
            raise ValueError(
                f"No phrases found with {feature_mode} filtering. "
                f"Try adjusting pmi_threshold or min_count."
            )

        if verbose:
            logger.info(f"Extracted {len(vocabulary)} phrases for {feature_mode} mode")

        metadata['vocabulary_size'] = len(vocabulary)

    elif feature_mode == 'hybrid':
        # ハイブリッドモード：低PMIと全フレーズの複合
        # ここでは簡略化のため、低PMIを優先する
        vocabulary = extract_pmi_filtered_phrases(
            texts,
            mode='low_pmi',
            pmi_threshold=pmi_threshold,
            min_count=min_count
        )
        if vocabulary is None:
            vocabulary = None  # フォールバックして全語彙を使う
        metadata['vocabulary_size'] = len(vocabulary) if vocabulary else 'all'

    # ステップ2: ベクトライザーを構築
    if feature_mode in ['tfidf', 'hybrid']:
        vectorizer = TfidfVectorizer(
            analyzer=analyzer,
            ngram_range=ngram_range,
            vocabulary=vocabulary,
            max_features=max_features,
            min_df=min_df,
            max_df=max_df
        )
        if verbose:
            logger.info(f"Using TfidfVectorizer with analyzer={analyzer}, ngram_range={ngram_range}")
    else:
        # low_pmi, high_pmiモード
        vectorizer = TfidfVectorizer(
            analyzer=analyzer,
            vocabulary=vocabulary,
            min_df=min_df,
            max_df=max_df
        )
        if verbose:
            logger.info(f"Using TfidfVectorizer with vocabulary restriction")

    # ステップ3: 行列を変換
    matrix_sparse = vectorizer.fit_transform(texts)
    matrix = matrix_sparse.toarray()

    # ステップ4: 特徴名を取得
    feature_names = vectorizer.get_feature_names_out().tolist()

    if verbose:
        logger.info(f"Document-term matrix shape: {matrix.shape}")
        logger.info(f"Number of features: {len(feature_names)}")
        sparsity = (matrix == 0).sum() / (matrix.shape[0] * matrix.shape[1])
        logger.info(f"Sparsity: {sparsity:.1%}")

    return matrix, vectorizer, feature_names, metadata


def fit_nmf_model(
    matrix: np.ndarray,
    n_topics: int = 10,
    nmf_init: str = 'nndsvd',
    nmf_max_iter: int = 1000,
    random_state: int = 42,
    verbose: int = 1
) -> Tuple[NMF, np.ndarray, np.ndarray]:
    """
    NMFモデルを行列に適用

    Parameters:
        matrix (np.ndarray): ドキュメント-ターム行列
        n_topics (int): トピック数
        nmf_init (str): NMF初期化方法
        nmf_max_iter (int): 最大反復回数
        random_state (int): ランダムシード
        verbose (int): ログレベル

    Returns:
        Tuple: (NMFモデル, W行列, H行列)
    """
    if verbose:
        logger.info(f"Fitting NMF model with {n_topics} topics")

    # nndsvd初期化は制約がある: n_components <= min(n_samples, n_features)
    # これを満たさない場合はランダム初期化にフォールバック
    n_samples, n_features = matrix.shape
    actual_init = nmf_init
    if nmf_init == 'nndsvd' and n_topics > min(n_samples, n_features):
        if verbose:
            logger.warning(
                f"nndsvd requires n_topics ({n_topics}) <= min(samples={n_samples}, "
                f"features={n_features}). Using 'random' initialization instead."
            )
        actual_init = 'random'

    nmf = NMF(
        n_components=n_topics,
        init=actual_init,
        max_iter=nmf_max_iter,
        random_state=random_state
    )

    W = nmf.fit_transform(matrix)  # Document-topic行列
    H = nmf.components_  # Topic-term行列

    if verbose:
        logger.info(f"NMF fitting complete")
        logger.info(f"W (Document-topic) shape: {W.shape}")
        logger.info(f"H (Topic-term) shape: {H.shape}")

    return nmf, W, H


def normalize_matrix_by_row(matrix: np.ndarray) -> np.ndarray:
    """
    行ごとに行列を正規化（各行の合計が1になるように）

    Parameters:
        matrix (np.ndarray): 入力行列

    Returns:
        np.ndarray: 正規化された行列
    """
    row_sums = matrix.sum(axis=1, keepdims=True)
    # ゼロ除算を避ける
    row_sums[row_sums == 0] = 1
    return matrix / row_sums


def get_top_terms_for_topic(
    topic_vector: np.ndarray,
    feature_names: List[str],
    n_terms: int = 15
) -> List[Tuple[str, float]]:
    """
    トピック分布から上位ターム を取得

    Parameters:
        topic_vector (np.ndarray): トピック-ターム ベクトル
        feature_names (List[str]): ターム名のリスト
        n_terms (int): 返す上位ターム数

    Returns:
        List[Tuple[str, float]]: (ターム、スコア) のタプルのリスト
    """
    # スコアが高い順にソート
    top_indices = np.argsort(topic_vector)[::-1][:n_terms]
    return [
        (feature_names[idx], float(topic_vector[idx]))
        for idx in top_indices
    ]


def calculate_distance_matrix(
    matrix: np.ndarray,
    metric: str = 'cosine'
) -> np.ndarray:
    """
    ドキュメント間の距離行列を計算

    Parameters:
        matrix (np.ndarray): ドキュメント-トピック行列
        metric (str): 距離計算方法（'cosine', 'euclidean', 'manhattan'）

    Returns:
        np.ndarray: 距離行列
    """
    from sklearn.metrics.pairwise import cosine_distances, euclidean_distances

    if metric == 'cosine':
        return cosine_distances(matrix)
    elif metric == 'euclidean':
        return euclidean_distances(matrix)
    elif metric == 'manhattan':
        from scipy.spatial.distance import pdist, squareform
        distances = pdist(matrix, metric='cityblock')
        return squareform(distances)
    else:
        raise ValueError(f"Unknown metric: {metric}")



"""
品質チェッカーのテスト
"""

import pytest
from japhrase import QualityChecker


class TestQualityChecker:
    """QualityChecker クラスのテスト"""

    def test_check_forbidden_phrases_found(self):
        """禁止ワードが見つかる"""
        text = "これは悪い表現です。とても悪いです。"
        checker = QualityChecker(text)
        result = checker.check_forbidden_phrases(['悪い', '最悪'])

        assert result is False  # エラーがある
        assert len(checker.errors) >= 1
        assert checker.errors[0]['type'] == 'forbidden_phrase'

    def test_check_forbidden_phrases_not_found(self):
        """禁止ワードが見つからない"""
        text = "これは良い表現です。とても良いです。"
        checker = QualityChecker(text)
        result = checker.check_forbidden_phrases(['悪い', '最悪'])

        assert result is True  # エラーなし
        assert len(checker.errors) == 0

    def test_check_required_keywords_sufficient(self):
        """必須キーワードが十分"""
        text = "機械学習は重要です。機械学習は有用です。機械学習は発展中です。"
        checker = QualityChecker(text)
        result = checker.check_required_keywords({'機械学習': 2})

        assert result is True  # 警告なし
        assert len(checker.warnings) == 0

    def test_check_required_keywords_insufficient(self):
        """必須キーワードが不足"""
        text = "機械学習は重要です。"
        checker = QualityChecker(text)
        result = checker.check_required_keywords({'機械学習': 3})

        assert result is False  # 警告あり
        assert len(checker.warnings) == 1
        assert checker.warnings[0]['type'] == 'missing_keyword'

    def test_check_spelling_consistency_found(self):
        """表記ゆれが見つかる"""
        text = "ユーザー登録とユーザ管理があります。"
        checker = QualityChecker(text)
        rules = {'ユーザー': ['ユーザ']}
        result = checker.check_spelling_consistency(rules)

        assert result is False  # エラーあり
        assert len(checker.errors) >= 1
        assert checker.errors[0]['type'] == 'spelling_inconsistency'

    def test_check_spelling_consistency_not_found(self):
        """表記ゆれがない"""
        text = "アルゴリズムの実装があります。"
        checker = QualityChecker(text)
        # algo は見つからない
        rules = {'アルゴリズム': ['algo']}
        result = checker.check_spelling_consistency(rules)

        assert result is True  # エラーなし
        assert len(checker.errors) == 0

    def test_check_length_limits_too_short(self):
        """文書が短すぎる"""
        text = "短い"
        checker = QualityChecker(text)
        result = checker.check_length_limits(min_length=100)

        assert result is False
        assert len(checker.errors) == 1
        assert checker.errors[0]['type'] == 'too_short'

    def test_check_length_limits_too_long(self):
        """文書が長すぎる"""
        text = "a" * 1000
        checker = QualityChecker(text)
        result = checker.check_length_limits(max_length=100)

        assert result is False
        assert len(checker.errors) == 1
        assert checker.errors[0]['type'] == 'too_long'

    def test_check_length_limits_ok(self):
        """文書の長さが適切"""
        text = "a" * 100
        checker = QualityChecker(text)
        result = checker.check_length_limits(min_length=50, max_length=150)

        assert result is True
        assert len(checker.errors) == 0

    def test_check_paragraph_structure_sufficient(self):
        """段落数が十分"""
        text = "段落1\n\n段落2\n\n段落3"
        checker = QualityChecker(text)
        result = checker.check_paragraph_structure(min_paragraphs=2)

        assert result is True
        assert len(checker.warnings) == 0

    def test_check_paragraph_structure_insufficient(self):
        """段落数が不足"""
        text = "段落1"
        checker = QualityChecker(text)
        result = checker.check_paragraph_structure(min_paragraphs=3)

        assert result is False
        assert len(checker.warnings) == 1
        assert checker.warnings[0]['type'] == 'few_paragraphs'

    def test_run_all_checks_success(self):
        """すべてのチェックに合格"""
        text = "これは良い文書です。\n\nとても良い文書です。\n\n本当に良い文書です。"
        config = {
            'check': {
                'forbidden_phrases': ['悪い'],
                'required_keywords': {'良い': 1},
                'spelling_rules': {'文書': ['ぶんしょ']},
                'min_length': 10,
                'max_length': 1000,
                'min_paragraphs': 2
            }
        }
        checker = QualityChecker(text, config)
        success, errors, warnings = checker.run_all_checks()

        assert success is True
        assert len(errors) == 0

    def test_run_all_checks_with_errors(self):
        """チェック失敗"""
        text = "これは悪い文書です。"
        config = {
            'check': {
                'forbidden_phrases': ['悪い'],
                'required_keywords': {'必須': 1},
            }
        }
        checker = QualityChecker(text, config)
        success, errors, warnings = checker.run_all_checks()

        assert success is False
        assert len(errors) >= 1

    def test_get_report(self):
        """レポート生成"""
        text = "テキスト"
        checker = QualityChecker(text)
        checker.errors.append({'message': 'エラーです'})
        checker.warnings.append({'message': '警告です'})

        report = checker.get_report()
        assert '品質チェック' in report
        assert 'エラーです' in report
        assert '警告です' in report
        assert len(report) > 0

    def test_case_insensitive_matching(self):
        """大文字小文字を区別しない検索"""
        text = "User and USER and user"
        checker = QualityChecker(text)
        result = checker.check_required_keywords({'user': 3})

        assert result is True  # すべてマッチ

    def test_empty_text(self):
        """空のテキスト"""
        text = ""
        checker = QualityChecker(text)
        result = checker.check_forbidden_phrases(['test'])

        assert result is True  # エラーなし
        assert len(checker.errors) == 0
